Texture2D DepthNormalBuffer : register(t0);
Texture2D RandomVecMap : register(t1);

SamplerState DepthSampler : register(s0);
SamplerState RandomVectorSampler : register(s1);


cbuffer Transforms : register(b0)
{
	matrix world;
	matrix view;
	matrix proj;
};

cbuffer OcclusionDataBuffer : register(b1)
{
	float4 Resolution; // x = width, y = height, x = 1/width, y = 1/height
	float4 Perspective; // x = fov, y = aspect, z = yn, w = xn
	float4 Frustum; // x = znear, y = zfar, z = znear/zfar, w = zfar/zfar
	float4 offsetVectors[14];
};

cbuffer OcclusionFactors : register(b2)
{
	float occlusionRadius;
	float occlusionFadeStart;
	float occlusionFadeEnd;
	float occlusionSurfaceEpsilon;
};

struct PSInput
{
	float4 pos : SV_POSITION;
	float2 tex : TEXCOORD0;
	float3 farPlane : TEXCOORD1;
};

float OcclusionCalculation(float distZ)
{
	float occlusion = 0.0f;

	if(distZ > occlusionSurfaceEpsilon)
	{
		float fadeLength = occlusionFadeEnd - occlusionFadeStart;

		occlusion = saturate((occlusionFadeEnd - distZ) / fadeLength);
	}

	return occlusion;
}

float4 main(in PSInput input) : SV_TARGET
{
	float4 testData;
	float4 normalDepth = DepthNormalBuffer.SampleLevel(DepthSampler, input.tex, 0.0f);
	float3 normal = normalDepth.xyz;
	float depth = normalDepth.w;
	int sampleCount = 14;

	float3 pixel = (depth / input.farPlane.z) * input.farPlane;

	float occlusionSum = 0.0f;

	float3 randomVector = 2.0f * RandomVecMap.SampleLevel(RandomVectorSampler, 4.0f * input.tex, 0.0f).rgb - 1.0f;

	[unroll]
	for (int i = 0; i < sampleCount; ++i)
	{
		// get the sample offset vector
		float3 offset = reflect(offsetVectors[i].xyz, randomVector);
		
		// flip the offset vecotr if it is behind the plane
		float flip = sign(dot(offset, normal));

		// sample a point near the pixel within the occlusion radius
		float3 q = pixel + flip * occlusionRadius * offset;

		// move the point to projection space
		float4 projQ = mul(float4(q, 1.0f), proj);
		projQ /= projQ.w;

		testData = float4(projQ.xyz, 1.0f);

		// find the nearest depth value along the ray
		float rayDepth = DepthNormalBuffer.SampleLevel(DepthSampler, projQ.xy, 0.0f).a;

		// reconstruct view space position
		float3 ray = (rayDepth / q.z) * q;

		// test whether r occludes p
		float distZ = pixel.z - ray.z;
		float distPixel = max(dot(normal, normalize(ray - pixel)), 0.0f);
		float occlusion = distPixel * OcclusionCalculation(distZ);
		occlusionSum += occlusion;
	}

	occlusionSum /= sampleCount;

	float access = 1.0f - occlusionSum;

	// sharpen the ao contrast to make the effect more dramatic
	return saturate(pow(access, 4.0f));
}
