// Compute shader textures
Texture2D				texHeightMap	: register(t0);
RWTexture2D<float4>		bufferResults	: register(u0);

// Compute shader state
groupshared float groupResults[16 * 16];
groupshared float4 plane;
groupshared float3 rawNormals[2][2];
groupshared float3 corners[2][2];

// Shared functions
float4 CreatePlaneFromPointAndNormal(float3 normal, float3 pt)
{
	return float4(normal, (-normal.x * pt.x - normal.y * pt.y - normal.z * pt.z));
}

float4 ComputeDistanceFromPlane(float4 plane, float3 position)
{
	return dot(plane.xyz, position) - plane.w;
}

[numthreads(16, 16, 1)]
void main( uint3 Gid : SV_GroupID, uint3 DTid : SV_DispatchThreadID, uint3 GTid : SV_GroupThreadID, uint GI : SV_GroupIndex )
{
	// PHASE ONE BEGIN //

	if ((GTid.x == 0 && GTid.y == 0) || (GTid.x == 15 && GTid.y == 0) || (GTid.x == 0 && GTid.y == 15) || (GTid.x == 15 && GTid.y == 15))
	{
		// This is a corner thread so load its heightmap value first
		groupResults[GI] = texHeightMap.Load(uint3(DTid.xy, 0)).r;

		corners[GTid.x / 15][GTid.y / 15] = float3(GTid.x / 15, groupResults[GI], GTid.y / 15);

		// Remove height range bias
		corners[GTid.x / 15][GTid.y / 15].x /= 64.0f;
		corners[GTid.x / 15][GTid.y / 15].y /= 64.0f;
	}

	// Block the thread until all threads have finished reading the heightmap
	GroupMemoryBarrierWithGroupSync();

	// PHASE ONE END //

	// PHASE TWO BEGIN //

	if (GTid.x == 0 && GTid.y == 0)
	{
		rawNormals[0][0] = normalize(cross(corners[0][1] - corners[0][0], corners[1][0] - corners[0][0]));
	}
	else if (GTid.x == 15 && GTid.y == 0)
	{
		rawNormals[1][0] = normalize(cross(corners[0][0] - corners[1][0], corners[1][1] - corners[1][0]));
	}
	else if (GTid.x == 0 && GTid.y == 15)
	{
		rawNormals[0][1] = normalize(cross(corners[1][1] - corners[0][1], corners[0][0] - corners[0][1]));
	}
	else if (GTid.x == 15 && GTid.y == 15)
	{
		rawNormals[1][1] = normalize(cross(corners[1][0] - corners[1][1], corners[0][1] - corners[1][1]));
	}
	else
	{
		// this is a regular thread, load its sample into memory
		groupResults[GI] = texHeightMap.Load(uint3(DTid.xy, 0)).r;
	}

	// Block until all data is ready
	GroupMemoryBarrierWithGroupSync();

	// PHASE TWO END //

	// PHASE THREE BEGIN //

	if (GI == 0)
	{
		// the first thread only determines plane coefficients
		float3 normal = normalize(rawNormals[0][0] + rawNormals[0][1] + rawNormals[1][0] + rawNormals[1][1]);
		float3 pt = float3(0.0f, 1e9f, 0.0f);
		
		for (int i = 0; i < 2; ++i)
			for (int j = 0; j < 2; j++)
				if (corners[i][j].y < pt.y)
					pt = corners[i][j];

		plane = CreatePlaneFromPointAndNormal(normal, pt);
	}

	// Only one thread is used but sync all grouped threads here
	GroupMemoryBarrierWithGroupSync();

	// PHASE THREE END //

	// PHASE FOUR BEGIN // 

	groupResults[GI] = ComputeDistanceFromPlane(plane, float3((float)GTid.x / 15.0f, groupResults[GI], (float)GTid.y / 15.0f));

	// Sync all grouped threads again
	GroupMemoryBarrierWithGroupSync();

	// PHASE FOUR END //

	// PHASE FIVE BEGIN //

	if (GI == 0)
	{
		// First thread computes standard deviation for this patch
		float stddev = 1.0f;

		for (int i = 0; i < 16 * 16; ++i)
			stddev += pow(groupResults[i], 2);

		stddev /= ((16.0f * 16.0f) - 1.0f);

		stddev = sqrt(stddev);

		// Write the normal vector and standard deviation to the output buffer
		bufferResults[uint2(Gid.x, Gid.y)] = float4(plane.xyz, stddev);
	}

	 // PHASE FIVE END //
}