#include "../res/shader_settings.hlsli"
//Textures
Texture2D normalMap : register(t0);
SamplerState anisoSampler : register(s0);

//Constants
cbuffer MaterialProperties
{
	float specularPower;
	float3 padding;
};

//Input/Output structures
struct PSInput
{
	float4 screenPosition : SV_Position;
	float2 texCoord : TEXCOORD0;
	float3 worldNormal : NORMAL;
	float3 worldTangent : TANGENT;
	float3 worldBinormal : BINORMAL;
	float3 worldPosition : POSITION0;
	float viewDepth : POSITION1;
};

struct PSOutput
{
	float4 gBuffer : SV_Target0;
	float4 position : SV_Target1;
};

//encode the normals for g-buffer packing
float2 SpheremapEncode(float3 normal)
{
	return normalize(normal.xy) * (sqrt(-normal.z * 0.5f + 0.5f));
}

//G-buffer pixel shader for pre-pass lighting
PSOutput main(in PSInput input, in uint coverageMask : SV_Coverage)
{
	PSOutput output;

	//normalize the tangent frame after interpolation
	float3x3 tangentFrame = float3x3(normalize(input.worldTangent), normalize(input.worldBinormal), normalize(input.worldNormal));

	//Sample the tangent space normal map and decompress
	float3 normalTS = normalMap.Sample(anisoSampler, input.texCoord).rgb;
	normalTS = normalize(normalTS * 2.0f - 1.0f);

	//convert to world space
	float3 worldNormal = mul(normalTS, tangentFrame);

	//encode the normals using sphere map transformation
	float2 encodedNormals = SpheremapEncode(worldNormal);
	
	//calculate the distance from the pixel to the view origin
	float distance = input.viewDepth;

	//calculate the coverage mask for this pixel
	const uint fullMask = (1 << (uint)SAMPLE_COUNT) - 1;
	float edgePixel = coverageMask != fullMask ? 1.0f : 0.0f;

	output.gBuffer = float4(encodedNormals.x, encodedNormals.y, edgePixel, specularPower);
	output.position = float4(input.worldPosition, distance);
	return output;
}