//Textures
Texture2D normalTexture : register(t0);
Texture2D positionTexture: register(t1);
SamplerState PointSampler : register(s0);

//Constants
cbuffer LightParams
{
	float4 lightPos;
	float4 lightColor;
	float4 lightDirection;
	float4 lightRange;
	float2 spotlightAngles;
	float lightType;
	float padding;
};

cbuffer CameraParams
{
	float3 cameraPos;
	float padding2;
};

cbuffer MatrixParams
{
	matrix invViewMatrix;
	matrix invProjMatrix;
};

#define DIRECTIONALLIGHT 0.0f
#define POINTLIGHT 1.0f
#define SPOTLIGHT 2.0f

//function for decoding normals from the sphere map
float3 SpheremapDecode(float2 encoded)
{
	float4 nn = float4(encoded, 1, -1);
	float l = dot(nn.xyz, -nn.xyw);
	nn.z = l;
	nn.xy *= sqrt(l);
	return nn.xyz * 2 + float3(0, 0, -1);
}

//function for extracting g-buffer attributes
void GetGBufferAttributes(in float2 screenPos, out float3 normal, out float3 position, out float specularPower, out float distance)
{
	//determine indices for sampling the texture from screen position
	int3 sampleIndices = int3(screenPos.xy, 0);
	float2 texCoords = screenPos;
	texCoords.x = (texCoords.x + 1.0f) / 2.0f;
	texCoords.y = (texCoords.y + 1.0f) / 2.0f;
	float4 normalTex = normalTexture.Load(sampleIndices);
	float2 encodedNormal = normalTex.xy;
	normal = SpheremapDecode(encodedNormal);
	//float3 distanceTex = normalTexture.Sample(PointSampler, texCoords).rgb;
	distance = normalTex.z;
	specularPower = normalTex.w;
	position = positionTexture.Load(sampleIndices).xyz;
}

//Calculate basic lighting for g-buffer texel
float4 CalcLighting(in float3 normal, in float3 position, in float specularPower)
{
	//calculate the diffuse term;
	float3 L = 0;
	float attenuation = 1.0f;
	if (lightType == POINTLIGHT || lightType == SPOTLIGHT)
	{
		//base the light vector on light position
		L = lightPos.xyz - position;
		//calculate attenuation based on distance from the light source
		float dist = length(L);
		attenuation = max(0, 1.0f - (dist / lightRange.x));

		L /= dist;

		L = normalize(L);
	}
	else if (lightType == DIRECTIONALLIGHT)
	{
		//light direction is explicit for directional lights
		L = -lightDirection.xyz;
	}

	if (lightType == SPOTLIGHT)
	{
		//Also add in the spotlight attenuation factor
		float3 L2 = lightDirection.xyz;
		float rho = dot(-L, L2);
		attenuation *= saturate((rho - spotlightAngles.y) / (spotlightAngles.x - spotlightAngles.y));
	}

	float nDotL = saturate(dot(normal, L));
	float3 diffuse = nDotL * lightColor * attenuation;

	//Calculate the specular amount
	float3 V = cameraPos - position;
	float3 H;
	if (lightType == DIRECTIONALLIGHT)
	{
		H = normalize(-L + V);
	}
	else
	{
		H = normalize(L + V);
	}
	float specular = pow(saturate(dot(normal, H)), specularPower) * attenuation * nDotL;

	//final value is diffuse rgb and mono specular
	return float4(diffuse, specular);
}

//Lighting pass pixel shader code
float4 main(in float4 screenPos : SV_POSITION) : SV_Target
{
	float3 normal;
	float3 position;
	float specularPower;
	float distance;

	//get g-buffer values
	GetGBufferAttributes(screenPos.xy, normal, position, specularPower, distance);

	/*
	//normalize the view ray and get the pixel position based on depth
	float3 viewPos = mul(screenPos, invProjMatrix).xyz;
	float3 viewRay = float3(viewPos.xy * (1000.0f / viewPos.z), 1000.0f);
	float3 viewPosition = viewRay * distance;
	position = mul(float4(viewPosition, 1.0f), invViewMatrix).xyz;
	*/
	return CalcLighting(normal, position, specularPower);
}