#include "../res/shaders/perlin_noise_func.hlsli"

Texture2D texHeightMap : register(t0);
SamplerState smpHeightMap : register(s0);

// Constants
cbuffer Transforms
{
	matrix invTposeWorld;
};

cbuffer SampleParams
{
	// xy = pixel dimensions
	// zw = geometry dimensions
	float4 heightMapDimensions;
};

struct DS_OUTPUT
{
	float4 position : SV_POSITION;
	float2 texCoord : TEXCOORD0;
	float3 normal : NORMAL;
	float3 color : COLOR;
};

// Output control point
struct HS_CONTROL_POINT_OUTPUT
{
	float3 position : CONTROL_POINT_POSITION;
	float2 texCoord : CONTROL_POINT_TEXCOORD;
	float2 noiseCoord : CONTROL_POINT_NORMAL;
};

// Output patch constant data.
struct HS_CONSTANT_DATA_OUTPUT
{
	float edgeTessellation[4]		: SV_TessFactor;
	float insideTessellation[2]		: SV_InsideTessFactor;
};

#define NUM_CONTROL_POINTS 4

float SampleHeightMap(float2 uv)
{
	// arbitrary bias to make the output more  A E S T H E T I C A L L Y  pleasing
	const float scale = 8.0f;
	return scale * texHeightMap.SampleLevel(smpHeightMap, uv, 0.0f).r;
}

float PerlinNoiseHeight(float2 position)
{
	return improved_perlin_noise(position.x, 0.0f, position.y);
}

float3 Sobel(float2 tc)
{
	float2 pxSz = float2(1.0f / heightMapDimensions.x, 1.0f / heightMapDimensions.y);

	// Compute the offsets
	float2 offset00 = tc + float2(-pxSz.x, -pxSz.y);
	float2 offset10 = tc + float2(0.0f, -pxSz.y);
	float2 offset20 = tc + float2(pxSz.x, -pxSz.y);

	float2 offset01 = tc + float2(-pxSz.x, 0.0f);
	float2 offset21 = tc + float2(pxSz.x, 0.0f);

	float2 offset02 = tc + float2(-pxSz.x, pxSz.y);
	float2 offset12 = tc + float2(0.0f, pxSz.y);
	float2 offset22 = tc + float2(pxSz.x, pxSz.y);

	// Get the eight samples surrounding the current pixel
	float height00 = SampleHeightMap(offset00);
	float height10 = SampleHeightMap(offset10);
	float height20 = SampleHeightMap(offset20);

	float height01 = SampleHeightMap(offset01);
	float height21 = SampleHeightMap(offset21);

	float height02 = SampleHeightMap(offset02);
	float height12 = SampleHeightMap(offset12);
	float height22 = SampleHeightMap(offset22);

	// Evaluate the sobel filters
	float Gx = height00 - height20 + 2.0f * height01 - 2.0f * height21 + height02 - height22;
	float Gy = height00 + 2.0f * height10 + height20 - height02 - 2.0f * height12 - height22;

	// Generate the missing Z
	float Gz = 0.01f * sqrt(max(0.0f, 1.0f - Gx * Gx - Gy * Gy));

	// Make sure the returned normal is of unit length
	return normalize(float3(2.0f * Gx, Gz, 2.0f * Gy));
}

[domain("quad")]
DS_OUTPUT main(
	HS_CONSTANT_DATA_OUTPUT input,
	float2 uv : SV_DomainLocation,
	const OutputPatch<HS_CONTROL_POINT_OUTPUT, NUM_CONTROL_POINTS> patch)
{
	DS_OUTPUT Output;

	// take the three world space coordinates and the interpolation values and determine interpolated position
	float3 finalVertexCoord = float3(0.0f, 0.0f, 0.0f);

	finalVertexCoord.xz = patch[0].position.xz * (1.0f - uv.x) * (1.0f - uv.y)
						+ patch[1].position.xz * uv.x * (1.0f - uv.y)
						+ patch[2].position.xz * (1.0f - uv.x) * uv.y
						+ patch[3].position.xz * uv.x * uv.y;


	float2 texCoord = patch[0].texCoord * (1.0f - uv.x) * (1.0f - uv.y)
					+ patch[1].texCoord * uv.x * (1.0f - uv.y)
					+ patch[2].texCoord * (1.0f - uv.x) * uv.y
					+ patch[3].texCoord * uv.x * uv.y;

	float2 noiseCoord = patch[0].noiseCoord * (1.0f - uv.x) * (1.0f - uv.y)
					  + patch[1].noiseCoord * uv.x * (1.0f - uv.y)
					  + patch[2].noiseCoord * (1.0f - uv.x) * uv.y
					  + patch[3].noiseCoord * uv.x * uv.y;

	// Determine height from the texture
	finalVertexCoord.y = SampleHeightMap(texCoord);

	// Determine height from perlin noise
	//finalVertexCoord.y = PerlinNoiseHeight(noiseCoord) * 8.0f;

	Output.texCoord = texCoord;

	Output.position = float4(finalVertexCoord, 1.0f);

	// Perform a sobel filter on the heightmap to determine an appropriate normal vector
	float3 normal = Sobel(texCoord);
	Output.normal = normalize(mul(float4(normal, 1.0f), invTposeWorld).xyz);
	
	// DEBUG SHADING
	float lod = (input.insideTessellation[0] - 5.0f) / (10.0f - 5.0f);

	// 0% = black
	// 33% = blue
	// 66% = green
	// 100% = red

	if (lod > 0.66f)
	{
		lod -= 0.66f;
		lod /= 0.34f;
	}
	else if (lod > 0.33f)
	{
		// High Detail
		lod -= 0.33f;
		lod /= 0.33f;
	}
	else
	{
		// Low Detail
		lod /= 0.33f;
	}

	Output.color = float3(lod, lod, lod);

	return Output;
}
