// LOD lookup buffer
Texture2D texLODLookup : register(t0);

// Constants
cbuffer Camera
{
	float3 cameraPos;
	float padding;
};

cbuffer MinMax
{
	float4 minMaxDistance;
	float4 minMaxLOD;
};

cbuffer SampleParams
{
	// xy = pixel dimensions
	// zw = geometry dimensions
	float4 heightMapDimensions;
};

// Input control point
struct VS_CONTROL_POINT_OUTPUT
{
	float3 position : WORLD_SPACE_CONTROL_POINT_POSITION;
	float2 texCoord : CONTROL_POINT_TEXCOORD;
	float2 noiseCoord : CONTROL_POINT_NORMAL;
};

// Output control point
struct HS_CONTROL_POINT_OUTPUT
{
	float3 position : CONTROL_POINT_POSITION;
	float2 texCoord : CONTROL_POINT_TEXCOORD;
	float2 noiseCoord : CONTROL_POINT_NORMAL;
};

// Output patch constant data.
struct HS_CONSTANT_DATA_OUTPUT
{
	float edgeTessellation[4]		: SV_TessFactor;
	float insideTessellation[2]		: SV_InsideTessFactor;
};

#define NUM_CONTROL_POINTS 12

// Shared functions
float3 ComputePatchMidPoint(float3 cp0, float3 cp1, float3 cp2, float3 cp3)
{
	return (cp0 + cp1 + cp2 + cp3) / 4.0f;
}

float ComputeScaledDistance(float3 from, float3 to)
{
	// Compute the raw distance from the camera to the midpoint of this patch
	float d = distance(from, to);

	// Scale this to be 0.0 (min dist) and 1.0 (max dist)
	return (d - minMaxDistance.x) / (minMaxDistance.y - minMaxDistance.x);
}

float ComputePatchLOD(float3 midPoint)
{
	// Compute the scaled distance
	float d = ComputeScaledDistance(cameraPos, midPoint);

	// Transform the 0.0-1.0 distance scale into the desired LOD
	return lerp(minMaxLOD.x, minMaxLOD.y, 1.0f - d);
}

float4 ReadLookup(uint2 idx)
{
	return texLODLookup.Load(uint3(idx, 0));
}

uint2 ComputeLookupIndex(uint patch, int xOffset, int zOffset)
{
	uint2 p;

	p.x = patch % (uint)heightMapDimensions.z;
	p.y = patch / (uint)heightMapDimensions.w;

	p.x = clamp(p.x + xOffset, 0, ((uint)heightMapDimensions.z) - 1);
	p.y = clamp(p.y + zOffset, 0, ((uint)heightMapDimensions.w) - 1);

	return p;
}

float ComputePatchLODUsingLookup(float3 midPoint, float4 lookup)
{
	// Compute scaled distance
	float d = ComputeScaledDistance(cameraPos.xyz, midPoint);

	// fudge factor to standard deviation for better results
	float sd = lookup.w / 0.75;

	float lod = sd * (1.0f - d);

	// transform this value to an actual LOD
	return lerp(minMaxLOD.x, minMaxLOD.y, clamp(lod, 0.0f, 1.0f));
}

// Patch Constant Function
HS_CONSTANT_DATA_OUTPUT CalcHSPatchConstants(
	InputPatch<VS_CONTROL_POINT_OUTPUT, NUM_CONTROL_POINTS> ip,
	uint PatchID : SV_PrimitiveID)
{
	HS_CONSTANT_DATA_OUTPUT Output;

	// Get the lookup values for these patches
	float4 tileLookup[] =
	{
		// Main quad
		ReadLookup(ComputeLookupIndex(PatchID, 0, 0)),
		// +x neighbour
		ReadLookup(ComputeLookupIndex(PatchID, 0, 1)),
		// +z neighbour
		ReadLookup(ComputeLookupIndex(PatchID, 1, 0)),
		// -x neighbour
		ReadLookup(ComputeLookupIndex(PatchID, 0, -1)),
		// -z neighbour
		ReadLookup(ComputeLookupIndex(PatchID, -1, 0))
	};
	
	// Determine the midpoint of the patch
	float3 midPoints[] =
	{
		// Main quad
		ComputePatchMidPoint(ip[0].position, ip[1].position, ip[2].position, ip[3].position),
		// +x neighbour
		ComputePatchMidPoint(ip[2].position, ip[3].position, ip[4].position, ip[5].position),
		// +z neighbour
		ComputePatchMidPoint(ip[1].position, ip[3].position, ip[6].position, ip[7].position),
		// -x neighbour
		ComputePatchMidPoint(ip[0].position, ip[1].position, ip[8].position, ip[9].position),
		// -z neighbour
		ComputePatchMidPoint(ip[0].position, ip[2].position, ip[10].position, ip[11].position)
	};
	
	// Determine the LOD for each patch
	float detail[] =
	{
		// Main quad
		ComputePatchLODUsingLookup(midPoints[0], tileLookup[0]),
		// +x neighbour
		ComputePatchLODUsingLookup(midPoints[1], tileLookup[1]),
		// +z neighbour
		ComputePatchLODUsingLookup(midPoints[2], tileLookup[2]),
		// -x neighbour
		ComputePatchLODUsingLookup(midPoints[3], tileLookup[3]),
		// -z neighbour
		ComputePatchLODUsingLookup(midPoints[4], tileLookup[4])
	};
	/*
	float detail[] =
	{
		// Main quad
		ComputePatchLOD(midPoints[0]),
		// +x neighbour
		ComputePatchLOD(midPoints[1]),
		// +z neighbour
		ComputePatchLOD(midPoints[2]),
		// -x neighbour
		ComputePatchLOD(midPoints[3]),
		// -z neighbour
		ComputePatchLOD(midPoints[4])
	};
	*/
	// The patch always has an interior matching the patch LOD
	Output.insideTessellation[0] = Output.insideTessellation[1] = detail[0];

	// Edges have more complex rules:
	// If the neighbour patch is of a lower LOD then pick that LOD as the edge
	// If the neighbour patch is higher LOD then keep this LOD and the neighbour will blend towards this
	Output.edgeTessellation[0] = min(detail[0], detail[4]);
	Output.edgeTessellation[1] = min(detail[0], detail[3]);
	Output.edgeTessellation[2] = min(detail[0], detail[2]);
	Output.edgeTessellation[3] = min(detail[0], detail[1]);

	return Output;
}

[domain("quad")]
[partitioning("fractional_odd")]
[outputtopology("triangle_cw")]
[outputcontrolpoints(4)]
[patchconstantfunc("CalcHSPatchConstants")]
HS_CONTROL_POINT_OUTPUT main( 
	InputPatch<VS_CONTROL_POINT_OUTPUT, NUM_CONTROL_POINTS> ip, 
	uint i : SV_OutputControlPointID)
{
	HS_CONTROL_POINT_OUTPUT Output;

	Output.position = ip[i].position;
	Output.texCoord = ip[i].texCoord;
	Output.noiseCoord = ip[i].noiseCoord;

	return Output;
}
