// Constants
cbuffer LightParams
{
	float4 lightPos;
	float4 lightColor;
	float4 lightDirection;
	float4 lightRange;
	float2 spotlightAngles;
	float lightType;
	float padding;
};

cbuffer MaterialParams
{
	float3 specularAlbedo;
	float specularPower;
};

cbuffer Camera
{
	float3 cameraPos;
	float padding2;
};

// I/O Structs
struct PSInput
{
	float4 position : SV_Position;
	float2 texCoord : TEXCOORD0;
	float3 normal : NORMAL;
	float4 tangent : TANGENT;
};

#define DIRECTIONALLIGHT 0.0f
#define POINTLIGHT 1.0f
#define SPOTLIGHT 2.0f

//Calculate basic lighting for g-buffer texel
float4 CalcLighting(in float3 normal, in float3 position)
{
	//calculate the diffuse term;
	float3 L = 0;
	float attenuation = 1.0f;
	if (lightType == POINTLIGHT || lightType == SPOTLIGHT)
	{
		//base the light vector on light position
		L = lightPos.xyz - position;
		//calculate attenuation based on distance from the light source
		float dist = length(L);
		attenuation = max(0, 1.0f - (dist / lightRange.x));

		L /= dist;

		L = normalize(L);
	}
	else if (lightType == DIRECTIONALLIGHT)
	{
		//light direction is explicit for directional lights
		L = -lightDirection.xyz;
	}

	if (lightType == SPOTLIGHT)
	{
		//Also add in the spotlight attenuation factor
		float3 L2 = lightDirection.xyz;
		float rho = dot(-L, L2);
		attenuation *= saturate((rho - spotlightAngles.y) / (spotlightAngles.x - spotlightAngles.y));
	}

	float nDotL = saturate(dot(normal, L));
	float3 diffuse = nDotL * lightColor.xyz * attenuation;

	//Calculate the specular amount
	float3 V = cameraPos - position;
	float3 H;
	if (lightType == DIRECTIONALLIGHT)
	{
		H = normalize(L + V);
	}
	else
	{
		H = normalize(L + V);
	}

	float specular = pow(saturate(dot(normal, H)), specularPower) * attenuation * nDotL;

	//final value is diffuse rgb and mono specular
	return float4(diffuse, specular);
}

float4 main(in PSInput input) : SV_TARGET
{
	float4 lighting = CalcLighting(input.normal, input.position);

	float3 diffuseAlbedo = min(0.75f, max(0.0f, dot(input.normal, float3(0.0f, 1.0f, 0.0f))));

	float3 diffuse = lighting.xyz * diffuseAlbedo;
	float3 specular = lighting.w * specularPower;

	float3 output = diffuse + specular;

	return float4(output, 1.0f);
}