// Structure that represents one fluid column's state
struct GridPoint
{
	float height;
	float4 flow;
};

// Input and output buffers
RWStructuredBuffer<GridPoint> NewWaterState : register(u0);
StructuredBuffer<GridPoint> CurrentWaterState : register(t0);

// Constant buffer parameters
cbuffer TimeParameters
{
	float4 timeFactors;
	float4 dispatchSize;
};

cbuffer SimulationParameters
{
	float timeStep;
	float pipeArea;
	float gravitation;
	float pipeLength;
	float fluidDensity;
	float columnArea;
	float dampingFactor;
	float padding;
};

// Simulation group size
#define size_x 16
#define size_y 16

// Group size with the extra perimeter for adjacent columns
#define padded_x ( 1 + size_x + 1)
#define padded_y ( 1 + size_y + 1)

// Shared memory for the padded group size
groupshared GridPoint loadedpoints[padded_x * padded_y];

[numthreads(padded_x, padded_y, 1)]
void main(uint3 Gid : SV_GroupID, uint3 DTid : SV_DispatchThreadID, uint3 GTid : SV_GroupThreadID, uint GI : SV_GroupIndex)
{
	// grid size
	int gridsize_x = dispatchSize.x;
	int gridsize_y = dispatchSize.y;

	// total texture size
	int totalsize_x = dispatchSize.z;
	int totalsize_y = dispatchSize.w;

	// Load groupshared memory with last frame's height and flow
	loadedpoints[GI].height = 0.0f;
	loadedpoints[GI].flow = float4(0.0f, 0.0f, 0.0f, 0.0f);

	// Calculate the thread's location in the buffer
	int3 location = int3(0, 0, 0);
	location.x = Gid.x * size_x + (GTid.x - 1);
	location.y = Gid.y * size_y + (GTid.y - 1);
	int textureIndex = location.x + location.y * totalsize_x;

	// Load the data into gsm
	loadedpoints[GI] = CurrentWaterState[textureIndex];

	// Sync all threads before moving on
	GroupMemoryBarrierWithGroupSync();

	// Initialize the flow variable to the last frames flow values
	float4 NewFlow = float4(0.0f, 0.0f, 0.0f, 0.0f);

	// Check the current thread isn't at the edge of the simulation
	if ((GTid.x < padded_x - 1) && (location.x < totalsize_x - 1))
	{
		NewFlow.x = (loadedpoints[GI + 1].height - loadedpoints[GI].height);

		if ((GTid.y < padded_y - 1) && (location.y < totalsize_y - 1))
		{
			NewFlow.y = (loadedpoints[(GI + 1) + padded_x].height - loadedpoints[GI].height);
		}
	}

	if ((GTid.y < padded_y - 1) && (location.y < totalsize_y - 1))
	{
		NewFlow.z = (loadedpoints[GI + padded_x].height - loadedpoints[GI].height);
		
		if ((GTid.x > 0) && (location.x > 0))
		{
			NewFlow.w = (loadedpoints[GI + padded_x - 1].height - loadedpoints[GI].height);
		}
	}

	float fAccelFactor = (min(timeFactors.x, timeStep) * pipeArea * gravitation) / (pipeLength * columnArea);

	// calculate the new flow and add the previous flow value
	NewFlow = (NewFlow * fAccelFactor + loadedpoints[GI].flow) * dampingFactor;

	// store the updated flow value in the group shared memory for other threads to access
	loadedpoints[GI].flow = NewFlow;

	// synchronize all threads before moving on
	GroupMemoryBarrierWithGroupSync();

	// Calculate the new height for each column
	// Current column out
	loadedpoints[GI].height = loadedpoints[GI].height + NewFlow.x + NewFlow.y + NewFlow.z + NewFlow.w;

	// left columns
	loadedpoints[GI].height = loadedpoints[GI].height - loadedpoints[GI - 1].flow.x;

	// upper left columns
	loadedpoints[GI].height = loadedpoints[GI].height - loadedpoints[GI - padded_x - 1].flow.y;

	// top columns
	loadedpoints[GI].height = loadedpoints[GI].height - loadedpoints[GI - padded_x].flow.z;

	// top right columns
	loadedpoints[GI].height = loadedpoints[GI].height - loadedpoints[GI - padded_x + 1].flow.w;

	// HACK TO STOP WATER GOING BELOW BASE TERRAIN LEVEL //
	if (loadedpoints[GI].height <= 0.0f)
		loadedpoints[GI].height = 0.0f;

	// Store the updated height and flow values only for the threads within the actual group
	if ((GTid.x > 0) && (GTid.x < padded_x - 1) && (GTid.y > 0) && (GTid.y < padded_y - 1))
	{
		NewWaterState[textureIndex] = loadedpoints[GI];
	}
}