// base shader.cpp
#include "baseshader.h"
#include "ResourceContainer.h"

BaseShader::BaseShader(ID3D11Device* device, HWND hwnd)
{
	m_device = device;
	m_hwnd = hwnd;
	m_layout = 0;
	m_blendState = 0;
	m_rasterizerState = 0;
	m_depthStencilState = 0;
	m_stencilRef = 1;
	m_blendFactor[0] = 0.0f;
	m_blendFactor[1] = 0.0f;
	m_blendFactor[2] = 0.0f;
	m_blendFactor[3] = 0.0f;
}


BaseShader::~BaseShader()
{
	// Shutdown the vertex and pixel shaders as well as the related objects.
	// Release the pixel shader.
	if (m_pixelShader)
	{
		m_pixelShader->Release();
		m_pixelShader = 0;
	}

	// Release the vertex shader.
	if (m_vertexShader)
	{
		m_vertexShader->Release();
		m_vertexShader = 0;
	}

	// Release the hull shader.
	if (m_hullShader)
	{
		m_hullShader->Release();
		m_hullShader = 0;
	}

	// Release the domain shader.
	if (m_domainShader)
	{
		m_domainShader->Release();
		m_domainShader = 0;
	}

	// Release the geometry shader.
	if (m_geometryShader)
	{
		m_geometryShader->Release();
		m_geometryShader = 0;
	}

	// Release the compute shader.
	if (m_computeShader)
	{
		m_computeShader->Release();
		m_computeShader = 0;
	}

	// Release the custom layout
	if (m_layout)
	{
		m_layout->Release();
		m_layout = 0;
	}

	// Release the custom blend state
	if (m_blendState)
	{
		m_blendState->Release();
		m_blendState = 0;
	}

	// Release the custom rasterizer state
	if (m_rasterizerState)
	{
		m_rasterizerState->Release();
		m_rasterizerState = 0;
	}

	// Release the custom depth stencil state
	if (m_depthStencilState)
	{
		m_depthStencilState->Release();
		m_depthStencilState = 0;
	}
}


void BaseShader::loadVertexShader(WCHAR* filename, LPCSTR entrypoint)
{

	HRESULT result;
	ID3DBlob* errorMessage;
	ID3DBlob* vertexShaderBuffer;	
	int layoutSize = (m_deferredShader) ? 1 : 4;
	layoutSize = (m_boneAnimating && !m_deferredShader) ? layoutSize + 2 : layoutSize;
	D3D11_INPUT_ELEMENT_DESC *polygonLayout = new D3D11_INPUT_ELEMENT_DESC[layoutSize];
	unsigned int numElements;
	
	// Initialize the pointers this function will use to null.
	errorMessage = 0;
	vertexShaderBuffer = 0;

	// Compile the vertex shader code.
	result = D3DCompileFromFile(filename, NULL, D3D_COMPILE_STANDARD_FILE_INCLUDE, entrypoint, "vs_5_0", D3DCOMPILE_ENABLE_STRICTNESS, 0, &vertexShaderBuffer, &errorMessage);
	if (FAILED(result))
	{
		// If the shader failed to compile it should have writen something to the error message.
		if (errorMessage)
		{
			OutputShaderErrorMessage(errorMessage, m_hwnd, filename);
		}
		// If there was nothing in the error message then it simply could not find the shader file itself.
		else
		{
			MessageBox(m_hwnd, filename, L"Missing Shader File", MB_OK);
		}
		exit(0);
	}

	// Create the vertex shader from the buffer.
	result = m_device->CreateVertexShader(vertexShaderBuffer->GetBufferPointer(), vertexShaderBuffer->GetBufferSize(), NULL, &m_vertexShader);
	if (FAILED(result))
	{
		//return false;
	}

	// Create the vertex input layout description.
	// This setup needs to match the VertexType stucture in the MeshClass and in the shader.
	polygonLayout[0].SemanticName = "POSITION";
	polygonLayout[0].SemanticIndex = 0;
	polygonLayout[0].Format = DXGI_FORMAT_R32G32B32_FLOAT;
	polygonLayout[0].InputSlot = 0;
	polygonLayout[0].AlignedByteOffset = 0;
	polygonLayout[0].InputSlotClass = D3D11_INPUT_PER_VERTEX_DATA;
	polygonLayout[0].InstanceDataStepRate = 0;

	if (!m_deferredShader)
	{
		polygonLayout[1].SemanticName = "TEXCOORD";
		polygonLayout[1].SemanticIndex = 0;
		polygonLayout[1].Format = DXGI_FORMAT_R32G32_FLOAT;
		polygonLayout[1].InputSlot = 0;
		polygonLayout[1].AlignedByteOffset = D3D11_APPEND_ALIGNED_ELEMENT;
		polygonLayout[1].InputSlotClass = D3D11_INPUT_PER_VERTEX_DATA;
		polygonLayout[1].InstanceDataStepRate = 0;

		polygonLayout[2].SemanticName = "NORMAL";
		polygonLayout[2].SemanticIndex = 0;
		polygonLayout[2].Format = DXGI_FORMAT_R32G32B32_FLOAT;
		polygonLayout[2].InputSlot = 0;
		polygonLayout[2].AlignedByteOffset = D3D11_APPEND_ALIGNED_ELEMENT;
		polygonLayout[2].InputSlotClass = D3D11_INPUT_PER_VERTEX_DATA;
		polygonLayout[2].InstanceDataStepRate = 0;

		polygonLayout[3].SemanticName = "TANGENT";
		polygonLayout[3].SemanticIndex = 0;
		polygonLayout[3].Format = DXGI_FORMAT_R32G32B32_FLOAT;
		polygonLayout[3].InputSlot = 0;
		polygonLayout[3].AlignedByteOffset = D3D11_APPEND_ALIGNED_ELEMENT;
		polygonLayout[3].InputSlotClass = D3D11_INPUT_PER_VERTEX_DATA;
		polygonLayout[3].InstanceDataStepRate = 0;

		if (m_boneAnimating)
		{
			polygonLayout[4].SemanticName = "BONE";
			polygonLayout[4].SemanticIndex = 0;
			polygonLayout[4].Format = DXGI_FORMAT_R32G32B32A32_UINT;
			polygonLayout[4].InputSlot = 0;
			polygonLayout[4].AlignedByteOffset = D3D11_APPEND_ALIGNED_ELEMENT;
			polygonLayout[4].InputSlotClass = D3D11_INPUT_PER_VERTEX_DATA;
			polygonLayout[4].InstanceDataStepRate = 0;

			polygonLayout[5].SemanticName = "WEIGHTS";
			polygonLayout[5].SemanticIndex = 0;
			polygonLayout[5].Format = DXGI_FORMAT_R32G32B32A32_FLOAT;
			polygonLayout[5].InputSlot = 0;
			polygonLayout[5].AlignedByteOffset = D3D11_APPEND_ALIGNED_ELEMENT;
			polygonLayout[5].InputSlotClass = D3D11_INPUT_PER_VERTEX_DATA;
			polygonLayout[5].InstanceDataStepRate = 0;
		}
	}
	// Get a count of the elements in the layout.
	numElements = sizeof(polygonLayout) / sizeof(polygonLayout[0]);

	// Create the vertex input layout.
	result = m_device->CreateInputLayout(polygonLayout, layoutSize, vertexShaderBuffer->GetBufferPointer(), vertexShaderBuffer->GetBufferSize(),
		&m_layout);
	if (FAILED(result))
	{
		//return false;
	}

	// Release the vertex shader buffer and pixel shader buffer since they are no longer needed.
	vertexShaderBuffer->Release();
	vertexShaderBuffer = 0;
}

void BaseShader::loadPixelShader(WCHAR* filename, LPCSTR entrypoint)
{
	HRESULT result;
	ID3DBlob* errorMessage;
	ID3DBlob* pixelShaderBuffer;
	
	// Compile the pixel shader code.
	result = D3DCompileFromFile(filename, NULL, D3D_COMPILE_STANDARD_FILE_INCLUDE, entrypoint, "ps_5_0", D3DCOMPILE_ENABLE_STRICTNESS, 0, &pixelShaderBuffer, &errorMessage);
	if (FAILED(result))
	{
		// If the shader failed to compile it should have writen something to the error message.
		if (errorMessage)
		{
			OutputShaderErrorMessage(errorMessage, m_hwnd, filename);
		}
		// If there was  nothing in the error message then it simply could not find the file itself.
		else
		{
			MessageBox(m_hwnd, filename, L"Missing Shader File", MB_OK);
		}

		exit(0);
	}

	// Create the pixel shader from the buffer.
	result = m_device->CreatePixelShader(pixelShaderBuffer->GetBufferPointer(), pixelShaderBuffer->GetBufferSize(), NULL, &m_pixelShader);
	
	pixelShaderBuffer->Release();
	pixelShaderBuffer = 0;
}

void BaseShader::loadHullShader(WCHAR* filename, LPCSTR entrypoint)
{
	HRESULT result;
	ID3D10Blob* errorMessage;
	ID3D10Blob* hullShaderBuffer;

	// Compile the hull shader code.
	result = D3DCompileFromFile(filename, NULL, D3D_COMPILE_STANDARD_FILE_INCLUDE, entrypoint, "hs_5_0", D3DCOMPILE_ENABLE_STRICTNESS, 0, &hullShaderBuffer, &errorMessage);
	if (FAILED(result))
	{
		// If the shader failed to compile it should have writen something to the error message.
		if (errorMessage)
		{
			OutputShaderErrorMessage(errorMessage, m_hwnd, filename);
		}
		// If there was nothing in the error message then it simply could not find the shader file itself.
		else
		{
			MessageBox(m_hwnd, filename, L"Missing Shader File", MB_OK);
		}

		exit(0);
	}

	// Create the hull shader from the buffer.
	result = m_device->CreateHullShader(hullShaderBuffer->GetBufferPointer(), hullShaderBuffer->GetBufferSize(), NULL, &m_hullShader);
	
	hullShaderBuffer->Release();
	hullShaderBuffer = 0;
}

void BaseShader::loadDomainShader(WCHAR* filename, LPCSTR entrypoint)
{
	HRESULT result;
	ID3D10Blob* errorMessage;
	ID3D10Blob* domainShaderBuffer;

	// Compile the domain shader code.
	result = D3DCompileFromFile(filename, NULL, D3D_COMPILE_STANDARD_FILE_INCLUDE, entrypoint, "ds_5_0", D3DCOMPILE_ENABLE_STRICTNESS, 0, &domainShaderBuffer, &errorMessage);
	if (FAILED(result))
	{
		// If the shader failed to compile it should have writen something to the error message.
		if (errorMessage)
		{
			OutputShaderErrorMessage(errorMessage, m_hwnd, filename);
		}
		// If there was nothing in the error message then it simply could not find the shader file itself.
		else
		{
			MessageBox(m_hwnd, filename, L"Missing Shader File", MB_OK);
		}

		exit(0);
	}

	// Create the domain shader from the buffer.
	result = m_device->CreateDomainShader(domainShaderBuffer->GetBufferPointer(), domainShaderBuffer->GetBufferSize(), NULL, &m_domainShader);
	
	domainShaderBuffer->Release();
	domainShaderBuffer = 0;
}

void BaseShader::loadGeometryShader(WCHAR* filename, LPCSTR entrypoint)
{
	HRESULT result;
	ID3D10Blob* errorMessage;
	ID3D10Blob* geometryShaderBuffer;

	// Compile the domain shader code.
	result = D3DCompileFromFile(filename, NULL, D3D_COMPILE_STANDARD_FILE_INCLUDE, entrypoint, "gs_5_0", D3DCOMPILE_ENABLE_STRICTNESS, 0, &geometryShaderBuffer, &errorMessage);
	if (FAILED(result))
	{
		// If the shader failed to compile it should have writen something to the error message.
		if (errorMessage)
		{
			OutputShaderErrorMessage(errorMessage, m_hwnd, filename);
		}
		// If there was nothing in the error message then it simply could not find the shader file itself.
		else
		{
			MessageBox(m_hwnd, filename, L"Missing Shader File", MB_OK);
		}

		exit(0);
	}

	// Create the domain shader from the buffer.
	m_device->CreateGeometryShader(geometryShaderBuffer->GetBufferPointer(), geometryShaderBuffer->GetBufferSize(), NULL, &m_geometryShader);

	geometryShaderBuffer->Release();
	geometryShaderBuffer = 0;
}

void BaseShader::loadGeometryShaderWithStreamOutput(WCHAR* filename, D3D11_SO_DECLARATION_ENTRY* soDesc, int soDescCount, int rasterizedStream, LPCSTR entrypoint)
{
	HRESULT result;
	ID3D10Blob* errorMessage;
	ID3D10Blob* geometryShaderBuffer;

	// Compile the geometry shader code.
	result = D3DCompileFromFile(filename, NULL, D3D_COMPILE_STANDARD_FILE_INCLUDE, entrypoint, "gs_5_0", D3DCOMPILE_ENABLE_STRICTNESS, 0, &geometryShaderBuffer, &errorMessage);
	if (FAILED(result))
	{
		// If the shader failed to compile it should have writen something to the error message.
		if (errorMessage)
		{
			OutputShaderErrorMessage(errorMessage, m_hwnd, filename);
		}
		// If there was nothing in the error message then it simply could not find the shader file itself.
		else
		{
			MessageBox(m_hwnd, filename, L"Missing Shader File", MB_OK);
		}

		exit(0);
	}

	// Create the geometry shader from the buffer
	result = m_device->CreateGeometryShaderWithStreamOutput(geometryShaderBuffer->GetBufferPointer(), geometryShaderBuffer->GetBufferSize(), soDesc, soDescCount, NULL, 0, rasterizedStream, NULL, &m_geometryShader);

	geometryShaderBuffer->Release();
	geometryShaderBuffer = 0;
}

void BaseShader::loadComputeShader(WCHAR* filename, LPCSTR entrypoint)
{
	HRESULT result;
	ID3D10Blob* errorMessage;
	ID3D10Blob* computeShaderBuffer;

	// Compile the compute shader code
	result = D3DCompileFromFile(filename, NULL, D3D_COMPILE_STANDARD_FILE_INCLUDE, entrypoint, "cs_5_0", D3DCOMPILE_ENABLE_STRICTNESS, 0, &computeShaderBuffer, &errorMessage);
	if(FAILED(result))
	{
		// If the shader failed to compile it should have writen something to the error message.
		if (errorMessage)
		{
			OutputShaderErrorMessage(errorMessage, m_hwnd, filename);
		}
		// If there was nothing in the error message then it simply could not find the shader file itself.
		else
		{
			MessageBox(m_hwnd, filename, L"Missing Shader File", MB_OK);
		}

		exit(0);
	}

	// Create the compute shader from the buffer
	m_device->CreateComputeShader(computeShaderBuffer->GetBufferPointer(), computeShaderBuffer->GetBufferSize(), NULL, &m_computeShader);

	computeShaderBuffer->Release();
	computeShaderBuffer = 0;
}

void BaseShader::OutputShaderErrorMessage(ID3D10Blob* errorMessage, HWND hwnd, WCHAR* shaderFilename)
{
	char* compileErrors;
	unsigned long bufferSize, i;
	ofstream fout;


	// Get a pointer to the error message text buffer.
	compileErrors = (char*)(errorMessage->GetBufferPointer());

	// Get the length of the message.
	bufferSize = errorMessage->GetBufferSize();

	// Open a file to write the error message to.
	fout.open("shader-error.txt");

	// Write out the error message.
	for (i = 0; i<bufferSize; i++)
	{
		fout << compileErrors[i];
	}

	// Close the file.
	fout.close();

	// Release the error message.
	errorMessage->Release();
	errorMessage = 0;

	// Pop a message up on the screen to notify the user to check the text file for compile errors.
	MessageBox(hwnd, L"Error compiling shader.  Check shader-error.txt for message.", shaderFilename, MB_OK);

	return;
}


void BaseShader::Render(ID3D11DeviceContext* deviceContext, int indexCount)
{
	// Set the vertex input layout.
	deviceContext->IASetInputLayout(m_layout);

	// Set the vertex and pixel shaders that will be used to render.
	deviceContext->VSSetShader(m_vertexShader, NULL, 0);
	deviceContext->PSSetShader(m_pixelShader, NULL, 0);
	
	// if Hull shader is not null then set HS and DS
	if (m_hullShader)
	{
		deviceContext->HSSetShader(m_hullShader, NULL, 0);
		deviceContext->DSSetShader(m_domainShader, NULL, 0);
	}
	else
	{
		deviceContext->HSSetShader(NULL, NULL, 0);
		deviceContext->DSSetShader(NULL, NULL, 0);
	}

	// if geometry shader is not null then set GS
	if (m_geometryShader)
	{
		deviceContext->GSSetShader(m_geometryShader, NULL, 0);
	}
	else
	{
		deviceContext->GSSetShader(NULL, NULL, 0);
	}

	// Render the triangle.
	deviceContext->DrawIndexed(indexCount, 0, 0);
}

void BaseShader::Render(ID3D11DeviceContext* deviceContext, D3D* d3d, void* resourceContainer)
{
	if (m_layout)
	{
		// if this effect uses a custom input layout set it in the pipeline
		deviceContext->IASetInputLayout(m_layout);
	}

	if (m_blendState)
	{
		// if this effect uses a custom blend state set it in the pipeline
		deviceContext->OMSetBlendState(m_blendState, m_blendFactor, 0xffffffff);
	}

	if (m_depthStencilState)
	{
		// if this effect uses a custom depth stencil state set it in the pipeline
		deviceContext->OMSetDepthStencilState(m_depthStencilState, m_stencilRef);
	}

	if (m_rasterizerState)
	{
		// if this effect uses a custom rasterizer state set it in the pipeline
		deviceContext->RSSetState(m_rasterizerState);
	}

	// Set the vertex and pixel shaders that will be used to render.
	if (m_vertexShader)
	{
		deviceContext->VSSetShader(m_vertexShader, NULL, 0);
	}
	else
	{
		deviceContext->VSSetShader(NULL, NULL, 0);
	}

	if (m_pixelShader)
	{
		deviceContext->PSSetShader(m_pixelShader, NULL, 0);
	}
	else
	{
		deviceContext->PSSetShader(NULL, NULL, 0);
	}
	
	// if Hull shader is not null then set HS and DS
	if (m_hullShader)
	{
		deviceContext->HSSetShader(m_hullShader, NULL, 0);
		deviceContext->DSSetShader(m_domainShader, NULL, 0);
	}
	else
	{
		deviceContext->HSSetShader(NULL, NULL, 0);
		deviceContext->DSSetShader(NULL, NULL, 0);
	}


	// if geometry shader is not null then set GS
	if (m_geometryShader)
	{
		deviceContext->GSSetShader(m_geometryShader, NULL, 0);
	}
	else
	{
		deviceContext->GSSetShader(NULL, NULL, 0);
	}

	// if the compute shader is not null the set CS
	if (m_computeShader)
	{
		deviceContext->CSSetShader(m_computeShader, NULL, 0);
		deviceContext->Dispatch(m_threadGroupCountX, m_threadGroupCountY, m_threadGroupCountZ);
		return;
	}
	else
	{
		deviceContext->CSSetShader(NULL, NULL, 0);
	}

	// Render the triangle.
	switch (m_renderType)
	{
	case INDEXED: 
		deviceContext->DrawIndexed(m_indexCount, 0, 0);
		break;
	case AUTO:
		deviceContext->DrawAuto();
		break;
	}
}

void BaseShader::AddLight(int lightIndex)
{
	m_renderLights.push_back(lightIndex);
}

void BaseShader::AddObject(SceneObject* object)
{
	m_renderObjects.push_back(object);
}

void BaseShader::AddBuffers(int bufferIndex)
{
	m_renderBuffers.push_back(bufferIndex);
}

void BaseShader::AddCamera(int cameraIndex)
{
	m_renderCameras.push_back(cameraIndex);
}

void BaseShader::AddUAV(int uavIndex)
{
	m_renderUAVs.push_back(uavIndex);
}

void BaseShader::AddTexture(int textureIndex)
{
	m_renderTextures.push_back(textureIndex);
}

void BaseShader::AddFloat(int floatIndex)
{
	m_renderFloats.push_back(floatIndex);
}

void BaseShader::RemoveLight(int lightIndex)
{
	for (int i = 0; i < m_renderLights.size(); ++i)
	{
		if (m_renderLights[i] == lightIndex)
		{
			m_renderLights.erase(m_renderLights.begin() + i);
			return;
		}
	}
}

void BaseShader::RemoveObject(int objectIndex)
{
	if (objectIndex < 0 || objectIndex >= m_renderObjects.size())
		return;
	m_renderObjects.erase(m_renderObjects.begin() + objectIndex);
}

void BaseShader::RemoveObject(SceneObject* object)
{
	for (int i =0; i<m_renderObjects.size(); ++i)
	{ 
		if (m_renderObjects[i] == object)
		{
			m_renderObjects.erase(m_renderObjects.begin() + i);
			return;
		}
	}
}

void BaseShader::RemoveObject(std::wstring& name)
{
	for (int i = 0; i < m_renderObjects.size(); ++i)
	{
		if (m_renderObjects[i]->GetName() == name)
		{
			m_renderObjects.erase(m_renderObjects.begin() + i);
			return;
		}
	}
}

void BaseShader::RemoveCamera(int cameraIndex)
{
	for (int i = 0; i < m_renderCameras.size();)
	{
		if (m_renderCameras[i] == cameraIndex)
		{
			m_renderCameras.erase(m_renderCameras.begin() + i);
			return;
		}
	}
}

void BaseShader::RemoveBuffers(int bufferIndex)
{
	for (int i = 0; i < m_renderBuffers.size(); ++i)
	{
		if (m_renderBuffers[i] == bufferIndex)
		{
			m_renderBuffers.erase(m_renderBuffers.begin() + i);
			return;
		}
	}
}

void BaseShader::RemoveUAV(int uavIndex)
{
	for (int i = 0; i < m_renderUAVs.size(); ++i)
	{
		if (m_renderUAVs[i] == uavIndex)
		{
			m_renderUAVs.erase(m_renderUAVs.begin() + i);
			return;
		}
	}
}

void BaseShader::RemoveTexture(int textureIndex)
{
	for (int i = 0; i < m_renderTextures.size(); ++i)
	{
		if (m_renderTextures[i] == textureIndex)
		{
			m_renderTextures.erase(m_renderTextures.begin() + i);
			return;
		}
	}
}

void BaseShader::RemoveFloat(int floatIndex)
{
	for (int i = 0; i < m_renderFloats.size(); ++i)
	{
		if (m_renderFloats[i] == floatIndex)
		{
			m_renderFloats.erase(m_renderFloats.begin() + i);
			return;
		}
	}
}