#include "AmbientOcclusionShader.h"
#include "../DXFramework/ResourceContainer.h"

AmbientOcclusionShader::AmbientOcclusionShader(ID3D11Device* device, HWND hwnd) : BaseShader(device, hwnd)
{
	InitShader(L"../res/shaders/ambient_occlusion_vs.hlsl", L"../res/shaders/ambient_occlusion_ps.hlsl");
}

AmbientOcclusionShader::~AmbientOcclusionShader()
{
	if (m_psOcclusionDataBuffer)
	{
		m_psOcclusionDataBuffer->Release();
		m_psOcclusionDataBuffer = 0;
	}

	if (m_psOcclusionFactorsBuffer)
	{
		m_psOcclusionFactorsBuffer->Release();
		m_psOcclusionFactorsBuffer = 0;
	}

	BaseShader::~BaseShader();
}

void AmbientOcclusionShader::InitShader(WCHAR* vsFilename, WCHAR* psFilename)
{
	loadVertexShader(vsFilename);
	loadPixelShader(psFilename);

	// generate the offset vectors here
	XMFLOAT4 offsetVectors[14] =
	{
		// 8 cube corners
		XMFLOAT4(1.0f, 1.0f, 1.0f, 0.0f),
		XMFLOAT4(-1.0f, -1.0f, -1.0f, 0.0f),
		XMFLOAT4(-1.0f, 1.0f, 1.0f, 0.0f),
		XMFLOAT4(1.0f, -1.0f, -1.0f, 0.0f),
		XMFLOAT4(1.0f, 1.0f, -1.0f, 0.0f),
		XMFLOAT4(-1.0f, -1.0f, 1.0f, 0.0f),
		XMFLOAT4(-1.0f, 1.0f, -1.0f, 0.0f),
		XMFLOAT4(1.0f, -1.0f, 1.0f, 0.0f),
		// 6 centers of cube face
		XMFLOAT4(-1.0f, 0.0f, 0.0f, 0.0f),
		XMFLOAT4(1.0f, 0.0f, 0.0f, 0.0f),
		XMFLOAT4(0.0f, -1.0f, 0.0f, 0.0f),
		XMFLOAT4(0.0f, 1.0f, 0.0f, 0.0f),
		XMFLOAT4(0.0f, 0.0f, -1.0f, 0.0f),
		XMFLOAT4(0.0f, 0.0f, 1.0f, 0.0f)
	};


	for (int i = 0; i < 14; ++i)
	{
		float s = randf(0.25f, 1.0f);
		m_offsets[i].x = offsetVectors[i].x * s;
		m_offsets[i].y = offsetVectors[i].y * s;
		m_offsets[i].z = offsetVectors[i].z * s;
		m_offsets[i].w = offsetVectors[i].w * s;
	}

	m_matrixBuffer = CreateConstantBuffer(m_device, sizeof(MatrixBufferType), true, true, NULL);
	m_psOcclusionDataBuffer = CreateConstantBuffer(m_device, sizeof(PSOcclusionDataBufferType), true, true, NULL);
	m_psOcclusionFactorsBuffer = CreateConstantBuffer(m_device, sizeof(PSOcclusionFactorsBufferType), true, true, NULL);
	m_vsFrustumDataBuffer = CreateConstantBuffer(m_device, sizeof(VSFrustumDataBufferType), true, true, NULL);

	// Initialise the two required sampler states
	D3D11_SAMPLER_DESC samplerDesc;
	samplerDesc.Filter = D3D11_FILTER_MIN_MAG_MIP_POINT;
	samplerDesc.AddressU = D3D11_TEXTURE_ADDRESS_BORDER;
	samplerDesc.AddressV = D3D11_TEXTURE_ADDRESS_BORDER;
	samplerDesc.AddressW = D3D11_TEXTURE_ADDRESS_BORDER;
	samplerDesc.MipLODBias = 0.0f;
	samplerDesc.MaxAnisotropy = 1;
	samplerDesc.ComparisonFunc = D3D11_COMPARISON_ALWAYS;
	samplerDesc.BorderColor[0] = 0;
	samplerDesc.BorderColor[1] = 0;
	samplerDesc.BorderColor[2] = 0;
	samplerDesc.BorderColor[3] = 1e5f;
	samplerDesc.MinLOD = 0;
	samplerDesc.MaxLOD = D3D11_FLOAT32_MAX;

	m_device->CreateSamplerState(&samplerDesc, &m_depthSampler);

	samplerDesc.AddressU = D3D11_TEXTURE_ADDRESS_WRAP;
	samplerDesc.AddressV = D3D11_TEXTURE_ADDRESS_WRAP;
	samplerDesc.AddressW = D3D11_TEXTURE_ADDRESS_WRAP;

	m_device->CreateSamplerState(&samplerDesc, &m_randomMapSampler);	
	
	// Create the depth disabled view required for this shader
	D3D11_DEPTH_STENCIL_DESC depthStencilDescription;
	depthStencilDescription.DepthEnable = false;
	depthStencilDescription.DepthWriteMask = D3D11_DEPTH_WRITE_MASK_ALL;
	depthStencilDescription.DepthFunc = D3D11_COMPARISON_LESS;
	depthStencilDescription.StencilEnable = false;
	depthStencilDescription.StencilReadMask = 0xFF;
	depthStencilDescription.StencilWriteMask = 0xFF;
	depthStencilDescription.FrontFace.StencilFailOp = D3D11_STENCIL_OP_KEEP;
	depthStencilDescription.FrontFace.StencilDepthFailOp = D3D11_STENCIL_OP_INCR;
	depthStencilDescription.FrontFace.StencilPassOp = D3D11_STENCIL_OP_KEEP;
	depthStencilDescription.FrontFace.StencilFunc = D3D11_COMPARISON_ALWAYS;
	depthStencilDescription.BackFace.StencilFailOp = D3D11_STENCIL_OP_KEEP;
	depthStencilDescription.BackFace.StencilDepthFailOp = D3D11_STENCIL_OP_DECR;
	depthStencilDescription.BackFace.StencilPassOp = D3D11_STENCIL_OP_KEEP;
	depthStencilDescription.BackFace.StencilFunc = D3D11_COMPARISON_ALWAYS;

	// Create the state using the device.
	//m_device->CreateDepthStencilState(&depthStencilDescription, &m_depthStencilState);
}

void AmbientOcclusionShader::Render(ID3D11DeviceContext* deviceContext, D3D* d3d, void* resourceContainer)
{
	HRESULT result;
	D3D11_MAPPED_SUBRESOURCE mappedResource;
	PSOcclusionDataBufferType* occlusionDataBuffer;
	PSOcclusionFactorsBufferType* occlusionFactorsBuffer;
	VSFrustumDataBufferType* frustumDataBuffer;
	MatrixBufferType* matrixBuffer;
	XMMATRIX world, view, proj;
	int bufferNumber;

	float x = (float)d3d->GetScreenWidth();
	float y = (float)d3d->GetScreenHeight();
	float zn = d3d->GetScreenNear();
	float zf = d3d->GetScreenFar();
	float zn_normalized = zn / zf;
	float zf_normalized = zf / zf;
	float aspect = x / y;
	float fov = d3d->GetFOV();
	float yn = 2.0f * zn_normalized * tan(fov / 2.0f);
	float xn = aspect * yn;

	// build the frustum corner data
	float halfHeight = zf * tan(0.5f * fov);
	float halfWidth = aspect * halfHeight;

	XMFLOAT4 frustumCorners[4] =
	{
		XMFLOAT4(-halfWidth, -halfHeight, zf, 0.0f),
		XMFLOAT4(-halfWidth, halfHeight, zf, 0.0f),
		XMFLOAT4(halfWidth, halfHeight, zf, 0.0f),
		XMFLOAT4(halfWidth, -halfHeight, zf, 0.0f)
	};

	ResourceContainer* container = reinterpret_cast<ResourceContainer*>(resourceContainer);

	world = m_renderObjects[0]->GetTransformMatrix();
	container->GetCamera(m_renderCameras[0])->GetViewMatrix(view);
	d3d->GetProjectionMatrix(proj);
	static const XMMATRIX T(
		0.5f, 0.0f, 0.0f, 0.0f,
		0.0f, -0.5f, 0.0f, 0.0f,
		0.0f, 0.0f, 1.0f, 0.0f,
		0.5f, 0.5f, 0.0f, 1.0f);

	proj = XMMatrixMultiply(proj, T);

	// Map the ambient occlusion texture as the output of the pixel shader
	container->GetBuffers(m_renderBuffers[1])->SetRenderTargets(NULL);
	container->GetBuffers(m_renderBuffers[1])->ClearRenderTargets(0.0f, 0.0f, 0.0f, 0.0f);

	// Send the frustum data to the vertex shader
	result = deviceContext->Map(m_vsFrustumDataBuffer, 0, D3D11_MAP_WRITE_DISCARD, 0, &mappedResource);
	frustumDataBuffer = (VSFrustumDataBufferType*)mappedResource.pData;
	frustumDataBuffer->frustumCorners[0] = frustumCorners[0];
	frustumDataBuffer->frustumCorners[1] = frustumCorners[1];
	frustumDataBuffer->frustumCorners[2] = frustumCorners[2];
	frustumDataBuffer->frustumCorners[3] = frustumCorners[3];
	deviceContext->Unmap(m_vsFrustumDataBuffer, 0);
	bufferNumber = 0;
	deviceContext->VSSetConstantBuffers(bufferNumber, 1, &m_vsFrustumDataBuffer);

	// Send the transform data to the pixel shader
	result = deviceContext->Map(m_matrixBuffer, 0, D3D11_MAP_WRITE_DISCARD, 0, &mappedResource);
	matrixBuffer = (MatrixBufferType*)mappedResource.pData;
	matrixBuffer->world = XMMatrixTranspose(world);
	matrixBuffer->view = XMMatrixTranspose(view);
	matrixBuffer->projection = XMMatrixTranspose(proj);
	deviceContext->Unmap(m_matrixBuffer, 0);
	bufferNumber = 0;
	deviceContext->PSSetConstantBuffers(bufferNumber, 1, &m_matrixBuffer);

	// Send the occlusion data to the pixel shader
	result = deviceContext->Map(m_psOcclusionDataBuffer, 0, D3D11_MAP_WRITE_DISCARD, 0, &mappedResource);
	occlusionDataBuffer = (PSOcclusionDataBufferType*)mappedResource.pData;
	occlusionDataBuffer->resolution = XMFLOAT4(x, y, 1.0f / x, 1.0f / y);
	occlusionDataBuffer->perspective = XMFLOAT4(fov, aspect, yn, xn);
	occlusionDataBuffer->frustum = XMFLOAT4(zn, zf, zn_normalized, zf_normalized);
	for (int i = 0; i < 14; i++)
	{
		occlusionDataBuffer->offsetVectors[i] = m_offsets[i];
	}
	//memcpy(occlusionDataBuffer->offsetVectors, m_offsets, sizeof(m_offsets));
	deviceContext->Unmap(m_psOcclusionDataBuffer, 0);
	bufferNumber = 1;
	deviceContext->PSSetConstantBuffers(bufferNumber, 1, &m_psOcclusionDataBuffer);

	// Sned the occlusion data to the pixel shader
	result = deviceContext->Map(m_psOcclusionFactorsBuffer, 0, D3D11_MAP_WRITE_DISCARD, 0, &mappedResource);
	occlusionFactorsBuffer = (PSOcclusionFactorsBufferType*)mappedResource.pData;
	occlusionFactorsBuffer->occlusionRadius = 0.75f;
	occlusionFactorsBuffer->occlusionFadeStart = 0.2f;
	occlusionFactorsBuffer->occlusionFadeEnd = 5.0f;
	occlusionFactorsBuffer->occlusionSurfaceEpsilon = 0.01f;
	deviceContext->Unmap(m_psOcclusionFactorsBuffer, 0);
	bufferNumber = 2;
	deviceContext->PSSetConstantBuffers(bufferNumber, 1, &m_psOcclusionFactorsBuffer);

	// Send the depth buffer data to the compute shader
	ID3D11ShaderResourceView* depthBuffer = container->GetBuffers(m_renderBuffers[0])->GetShaderResourceView(0);
	deviceContext->PSSetShaderResources(0, 1, &depthBuffer);

	// Send the random vector map data to the pixel shader
	ID3D11ShaderResourceView* randomVecs = container->GetTexture(m_renderTextures[0])->GetTexture();
	deviceContext->PSSetShaderResources(1, 1, &randomVecs);

	// Set the two samplers in the pixel shader
	deviceContext->PSSetSamplers(0, 1, &m_depthSampler);
	deviceContext->PSSetSamplers(1, 1, &m_randomMapSampler);

	for (auto mesh : m_renderObjects[0]->GetMeshes())
	{
		m_indexCount = container->GetMesh(mesh)->GetIndexCount();
		container->GetMesh(mesh)->SendData(deviceContext);

		BaseShader::Render(deviceContext, d3d, resourceContainer);
	}
}