#include "ComputeAmbientOcclusionShader.h"
#include "../DXFramework/ResourceContainer.h"

ComputeAmbientOcclusionShader::ComputeAmbientOcclusionShader(ID3D11Device* device, HWND hwnd) : BaseShader(device, hwnd)
{
	InitShader(L"compute_ambient_occlusion_vs.hlsl", L"compute_ambient_occlusion_ps.hlsl");
}

ComputeAmbientOcclusionShader::~ComputeAmbientOcclusionShader()
{
	if (m_csOcclusionDataBuffer)
	{
		m_csOcclusionDataBuffer->Release();
		m_csOcclusionDataBuffer = 0;
	}

	if (m_csOcclusionFactorsBuffer)
	{
		m_csOcclusionFactorsBuffer->Release();
		m_csOcclusionFactorsBuffer = 0;
	}

	if (m_matrixBuffer)
	{
		m_matrixBuffer->Release();
		m_matrixBuffer = 0;
	}

	BaseShader::~BaseShader();
}

void ComputeAmbientOcclusionShader::InitShader(WCHAR* vsFilename, WCHAR* psFilename)
{
	loadVertexShader(vsFilename);
	loadPixelShader(psFilename);
	loadComputeShader(L"compute_ambient_occlusion_cs.hlsl");

	m_csOcclusionDataBuffer = CreateConstantBuffer(m_device, sizeof(CSOcclusionDataBufferType), true, true, NULL);
	m_csOcclusionFactorsBuffer = CreateConstantBuffer(m_device, sizeof(CSOcclusionFactorsBufferType), true, true, NULL);
	m_matrixBuffer = CreateConstantBuffer(m_device, sizeof(MatrixBufferType), true, true, NULL);
	
	// Initialise the two required sampler states
	D3D11_SAMPLER_DESC samplerDesc;
	samplerDesc.Filter = D3D11_FILTER_MIN_MAG_MIP_POINT;
	samplerDesc.AddressU = D3D11_TEXTURE_ADDRESS_CLAMP;
	samplerDesc.AddressV = D3D11_TEXTURE_ADDRESS_CLAMP;
	samplerDesc.AddressW = D3D11_TEXTURE_ADDRESS_CLAMP;
	samplerDesc.MipLODBias = 0.0f;
	samplerDesc.MaxAnisotropy = 1;
	samplerDesc.ComparisonFunc = D3D11_COMPARISON_ALWAYS;
	samplerDesc.BorderColor[0] = 0;
	samplerDesc.BorderColor[1] = 0;
	samplerDesc.BorderColor[2] = 0;
	samplerDesc.BorderColor[3] = 0;
	samplerDesc.MinLOD = 0;
	samplerDesc.MaxLOD = D3D11_FLOAT32_MAX;

	m_device->CreateSamplerState(&samplerDesc, &m_sampleState);
}

void ComputeAmbientOcclusionShader::Render(ID3D11DeviceContext* deviceContext, D3D* d3d, void* resourceContainer)
{
	HRESULT result;
	D3D11_MAPPED_SUBRESOURCE mappedResource;
	CSOcclusionDataBufferType* occlusionDataBuffer;
	CSOcclusionFactorsBufferType* occlusionFactorsBuffer;
	MatrixBufferType* matrixBuffer;
	XMMATRIX world, view, proj;
	int bufferNumber;

	float x = (float)d3d->GetScreenWidth();
	float y = (float)d3d->GetScreenHeight();
	float zn = d3d->GetScreenNear();
	float zf = d3d->GetScreenFar();
	float zn_normalized = zn / zf;
	float zf_normalized = zf / zf;
	float aspect = x / y;
	float fov = d3d->GetFOV();
	float yn = 2.0f * zn_normalized * tan(fov / 2.0f);
	float xn = aspect * yn;

	ResourceContainer* container = reinterpret_cast<ResourceContainer*>(resourceContainer);

	container->GetCamera(m_renderCameras[0])->GetViewMatrix(view);
	d3d->GetProjectionMatrix(proj);

	// Map the ambient occlusion texture as the output of the pixel shader
	container->GetBuffers(L"AMBIENT_OCCLUSION_TEXTURE")->SetRenderTargets(NULL);
	container->GetBuffers(L"AMBIENT_OCCLUSION_TEXTURE")->ClearRenderTargets(0.0f, 0.0f, 0.0f, 0.0f);

	// Send the occlusion data to the pixel shader
	result = deviceContext->Map(m_csOcclusionDataBuffer, 0, D3D11_MAP_WRITE_DISCARD, 0, &mappedResource);
	occlusionDataBuffer = (CSOcclusionDataBufferType*)mappedResource.pData;
	occlusionDataBuffer->resolution = XMFLOAT4(x, y, 1.0f / x, 1.0f / y);
	occlusionDataBuffer->perspective = XMFLOAT4(fov, aspect, yn, xn);

	deviceContext->Unmap(m_csOcclusionDataBuffer, 0);
	bufferNumber = 0;
	deviceContext->CSSetConstantBuffers(bufferNumber, 1, &m_csOcclusionDataBuffer);

	// Sned the occlusion data to the pixel shader
	result = deviceContext->Map(m_csOcclusionFactorsBuffer, 0, D3D11_MAP_WRITE_DISCARD, 0, &mappedResource);
	occlusionFactorsBuffer = (CSOcclusionFactorsBufferType*)mappedResource.pData;
	occlusionFactorsBuffer->occlusionRadius = 0.5f;
	occlusionFactorsBuffer->occlusionFadeStart = 0.2f;
	occlusionFactorsBuffer->occlusionFadeEnd = 2.0f;
	occlusionFactorsBuffer->occlusionSurfaceEpsilon = 0.05f;
	deviceContext->Unmap(m_csOcclusionFactorsBuffer, 0);
	bufferNumber = 1;
	deviceContext->CSSetConstantBuffers(bufferNumber, 1, &m_csOcclusionFactorsBuffer);

	// Send the depth buffer data to the compute shader
	ID3D11ShaderResourceView* depthBuffer = container->GetBuffers(m_renderBuffers[0])->GetShaderResourceView(0);
	deviceContext->CSSetShaderResources(0, 1, &depthBuffer);

	// Send the ambient occlusion uav to the compute shader
	ID3D11UnorderedAccessView* ambientBuffer = container->GetUAV(L"AMBIENT_OCCLUSION_UAV")->GetUAV();
	UINT initialCounts[] = { 0 };
	deviceContext->CSSetUnorderedAccessViews(0, 1, &ambientBuffer, initialCounts);

	// Set the compute shader sampler
	deviceContext->CSSetSamplers(0, 1, &m_sampleState);

	m_threadGroupCountX = x / 32;
	m_threadGroupCountY = y / 32;
	m_threadGroupCountZ = 1;

	deviceContext->CSSetShader(m_computeShader, 0, 0);
	deviceContext->Dispatch(m_threadGroupCountX, m_threadGroupCountY, m_threadGroupCountZ);

	// Unset the ambient occlusion uav so it can be used by the pixel shader
	ID3D11UnorderedAccessView* nullViews[] = { NULL };
	deviceContext->CSSetUnorderedAccessViews(0, 1, nullViews, initialCounts);

	// Set the ambient occlusion texture as the output of the pixel shader
	container->GetBuffers(L"AMBIENT_OCCLUSION_TEXTURE")->SetRenderTargets();
	container->GetBuffers(L"AMBIENT_OCCLUSION_TEXTURE")->ClearRenderTargets(0.0f, 0.0f, 0.0f, 0.0f);

	// Set the ambient occlusion UAV as input to the pixel shader
	ID3D11ShaderResourceView* ambientTexture = container->GetUAV(L"AMBIENT_OCCLUSION_UAV")->GetSRV();
	deviceContext->PSSetShaderResources(0, 1, &ambientTexture);


	deviceContext->VSSetShader(m_vertexShader, 0, 0);
	deviceContext->PSSetShader(m_pixelShader, 0, 0);

	for (auto object : m_renderObjects)
	{
		world = object->GetTransformMatrix();

		// Send the transform data to the pixel shader
		result = deviceContext->Map(m_matrixBuffer, 0, D3D11_MAP_WRITE_DISCARD, 0, &mappedResource);
		matrixBuffer = (MatrixBufferType*)mappedResource.pData;
		matrixBuffer->world = XMMatrixTranspose(world);
		matrixBuffer->view = XMMatrixTranspose(view);
		matrixBuffer->projection = XMMatrixTranspose(proj);
		deviceContext->Unmap(m_matrixBuffer, 0);
		bufferNumber = 0;
		deviceContext->PSSetConstantBuffers(bufferNumber, 1, &m_matrixBuffer);

		// Send the matrix data to the vertex shader 
		for (auto mesh : object->GetMeshes())
		{
			container->GetMesh(mesh)->SendData(deviceContext);
			m_indexCount = container->GetMesh(mesh)->GetIndexCount();
			m_renderType = container->GetMesh(mesh)->GetRenderType();

			// Render the triangle.
			switch (m_renderType)
			{
			case INDEXED:
				deviceContext->DrawIndexed(m_indexCount, 0, 0);
				break;
			case AUTO:
				deviceContext->DrawAuto();
				break;
			}
		}
	}
}