#include "DeferredGBufferShader.h"

DeferredGBufferShader::DeferredGBufferShader(ID3D11Device* device, HWND hwnd) : BaseShader(device, hwnd)
{
	InitShader(L"deferred_g_vs.hlsl", L"deferred_g_ps.hlsl");
}

DeferredGBufferShader::~DeferredGBufferShader()
{
	if (m_matrixBuffer)
	{
		m_matrixBuffer->Release();
		m_matrixBuffer = nullptr;
	}

	if (m_sampleState)
	{
		m_sampleState->Release();
		m_sampleState = nullptr;
	}

	if (m_materialBuffer)
	{
		m_materialBuffer->Release();
		m_materialBuffer = nullptr;
	}

	BaseShader::~BaseShader();

}
void DeferredGBufferShader::InitShader(WCHAR* vsFilename, WCHAR* psFilename)
{
	D3D11_SAMPLER_DESC samplerDesc;

	loadVertexShader(vsFilename);
	loadPixelShader(psFilename);

	//Create the constant buffer to access the vertex shader constant buffer
	m_matrixBuffer = CreateConstantBuffer(m_device, sizeof(MatrixBufferType), true, true, NULL);

	//Create the material constant buffer
	m_materialBuffer = CreateConstantBuffer(m_device, sizeof(MaterialBufferType), true, true, NULL);

	// Create a texture sampler state description.
	samplerDesc.Filter = D3D11_FILTER_MIN_MAG_MIP_LINEAR;
	samplerDesc.AddressU = D3D11_TEXTURE_ADDRESS_WRAP;
	samplerDesc.AddressV = D3D11_TEXTURE_ADDRESS_WRAP;
	samplerDesc.AddressW = D3D11_TEXTURE_ADDRESS_WRAP;
	samplerDesc.MipLODBias = 0.0f;
	samplerDesc.MaxAnisotropy = 1;
	samplerDesc.ComparisonFunc = D3D11_COMPARISON_ALWAYS;
	samplerDesc.BorderColor[0] = 0;
	samplerDesc.BorderColor[1] = 0;
	samplerDesc.BorderColor[2] = 0;
	samplerDesc.BorderColor[3] = 0;
	samplerDesc.MinLOD = 0;
	samplerDesc.MaxLOD = D3D11_FLOAT32_MAX;

	// Create the texture sampler state.
	m_device->CreateSamplerState(&samplerDesc, &m_sampleState);
}

void DeferredGBufferShader::SetShaderParameters(ID3D11DeviceContext* deviceContext, const XMMATRIX &world, const XMMATRIX &view, const XMMATRIX &projection, ID3D11ShaderResourceView* texture, ID3D11ShaderResourceView* normalMap)
{
	HRESULT result;
	D3D11_MAPPED_SUBRESOURCE mappedResource;
	MatrixBufferType* dataPtr;
	MaterialBufferType* matPtr;
	unsigned int bufferNumber;
	XMMATRIX tworld, tview, tproj;

	// Transpose the matrices to prepare them for the shader.
	tworld = XMMatrixTranspose(world);
	tview = XMMatrixTranspose(view);
	tproj = XMMatrixTranspose(projection);

	// Lock the constant buffer so it can be written to.
	result = deviceContext->Map(m_matrixBuffer, 0, D3D11_MAP_WRITE_DISCARD, 0, &mappedResource);

	// Get a pointer to the data in the constant buffer.
	dataPtr = (MatrixBufferType*)mappedResource.pData;

	// Copy the matrices into the constant buffer.
	dataPtr->world = tworld;// worldMatrix;
	dataPtr->view = tview;
	dataPtr->projection = tproj;

	// Unlock the constant buffer.
	deviceContext->Unmap(m_matrixBuffer, 0);

	// Set the position of the constant buffer in the vertex shader.
	bufferNumber = 0;

	// Now set the constant buffer in the vertex shader with the updated values.
	deviceContext->VSSetConstantBuffers(bufferNumber, 1, &m_matrixBuffer);

	// Send material data to pixel shader
	deviceContext->Map(m_materialBuffer, 0, D3D11_MAP_WRITE_DISCARD, 0, &mappedResource);

	//Get a pointer to the constant buffer data
	matPtr = (MaterialBufferType*)mappedResource.pData;

	//Copy the material data into the constant buffer
	matPtr->specularAlbedo = XMFLOAT3(1.0f, 1.0f, 1.0f);
	matPtr->specularPower = 25.0f;

	//Unlcok the constant buffer
	deviceContext->Unmap(m_materialBuffer, 0);

	bufferNumber = 0;

	//Set the constant buffer in the pixel shader
	deviceContext->PSSetConstantBuffers(bufferNumber, 1, &m_materialBuffer);

	ID3D11ShaderResourceView* textures[2];
	textures[0] = texture;
	textures[1] = normalMap;

	// Set shader texture resource in the pixel shader.
	deviceContext->PSSetShaderResources(0, 2, textures);
}


void DeferredGBufferShader::Render(ID3D11DeviceContext* deviceContext, int indexCount)
{
	// Set the sampler state in the pixel shader.
	deviceContext->PSSetSamplers(0, 1, &m_sampleState);
	
	BaseShader::Render(deviceContext, indexCount);
}