#include "DeferredShader.h"

DeferredShader::DeferredShader(ID3D11Device* device, HWND hwnd) : BaseShader(device, hwnd)
{
	m_deferredShader = true;
	InitShader(L"deferred_light_vs.hlsl", L"deferred_light_ps.hlsl");
}


DeferredShader::~DeferredShader()
{
	// Release the sampler state.
	if (m_sampleState)
	{
		m_sampleState->Release();
		m_sampleState = 0;
	}
	
	// Release the layout.
	if (m_layout)
	{
		m_layout->Release();
		m_layout = 0;
	}

	// Release the light constant buffer.
	if (m_lightBuffer)
	{
		m_lightBuffer->Release();
		m_lightBuffer = 0;
	}

	//Release the camera constant buffer
	if (m_cameraBuffer)
	{
		m_cameraBuffer->Release();
		m_cameraBuffer = 0;
	}

	//Release base shader components
	BaseShader::~BaseShader();
}


void DeferredShader::InitShader(WCHAR* vsFilename, WCHAR* psFilename)
{
	HRESULT result;
	D3D11_SAMPLER_DESC samplerDesc;

	loadVertexShader(vsFilename);

	loadPixelShader(psFilename);

	// Create a texture sampler state description.
	samplerDesc.Filter = D3D11_FILTER_MIN_MAG_MIP_LINEAR;
	samplerDesc.AddressU = D3D11_TEXTURE_ADDRESS_CLAMP;
	samplerDesc.AddressV = D3D11_TEXTURE_ADDRESS_CLAMP;
	samplerDesc.AddressW = D3D11_TEXTURE_ADDRESS_CLAMP;
	samplerDesc.MipLODBias = 0.0f;
	samplerDesc.MaxAnisotropy = 1;
	samplerDesc.ComparisonFunc = D3D11_COMPARISON_ALWAYS;
	samplerDesc.BorderColor[0] = 0;
	samplerDesc.BorderColor[1] = 0;
	samplerDesc.BorderColor[2] = 0;
	samplerDesc.BorderColor[3] = 0;
	samplerDesc.MinLOD = 0;
	samplerDesc.MaxLOD = D3D11_FLOAT32_MAX;

	// Create the texture sampler state.
	result = m_device->CreateSamplerState(&samplerDesc, &m_sampleState);
	if (FAILED(result))
	{
		printf("Sample State creation failed.");
	}

	//Create the light constant buffer
	m_lightBuffer = CreateConstantBuffer(m_device, sizeof(LightBufferType), true, true, NULL);

	//Create the camera constant buffer
	m_cameraBuffer = CreateConstantBuffer(m_device, sizeof(CameraBufferType), true, true, NULL);

}

void DeferredShader::SetShaderParameters(ID3D11DeviceContext* deviceContext, ID3D11ShaderResourceView* normalTexture, ID3D11ShaderResourceView* diffuseTexture, ID3D11ShaderResourceView* specularTexture, ID3D11ShaderResourceView* positionTexture, Light* light, const XMFLOAT3 &cameraPos)
{
	HRESULT result;
	D3D11_MAPPED_SUBRESOURCE mappedResource;
	LightBufferType* lightPtr;
	CameraBufferType* cameraPtr;
	unsigned int bufferNumber;

	// Send light data to pixel shader
	deviceContext->Map(m_lightBuffer, 0, D3D11_MAP_WRITE_DISCARD, 0, &mappedResource);
	lightPtr = (LightBufferType*)mappedResource.pData;
	XMFLOAT3 pos = light->GetPosition();
	lightPtr->lightPos = XMFLOAT4(pos.x, pos.y, pos.z, 1.0f);
	lightPtr->lightColor = light->GetDiffuseColour();
	XMFLOAT3 dir = light->GetDirection();
	lightPtr->lightDirection = XMFLOAT4(dir.x, dir.y, dir.z, 1.0f);
	lightPtr->lightRange = XMFLOAT4(25.0f, 25.0f, 25.0f, 25.0f);
	lightPtr->spotlightAngles = XMFLOAT2(45.0f, 45.0f);
	lightPtr->lightType = light->GetLightType();
	lightPtr->padding = 0.0f;

	deviceContext->Unmap(m_lightBuffer, 0);
	bufferNumber = 0;
	deviceContext->PSSetConstantBuffers(bufferNumber, 1, &m_lightBuffer);

	//Lock the camera constant buffer so it can be written to
	result = deviceContext->Map(m_cameraBuffer, 0, D3D11_MAP_WRITE_DISCARD, 0, &mappedResource);

	//Get a pointer to the constant buffer data
	cameraPtr = (CameraBufferType*)mappedResource.pData;

	//Copy the camera position into the constant buffer
	cameraPtr->cameraPos = cameraPos;
	cameraPtr->padding = 1.0f;

	//Unlcok the constant buffer
	deviceContext->Unmap(m_cameraBuffer, 0);

	bufferNumber = 1;

	//Set the constant buffer in the pixel shader
	deviceContext->PSSetConstantBuffers(bufferNumber, 1, &m_cameraBuffer);

	// Set shader texture resource in the pixel shader.
	deviceContext->PSSetShaderResources(0, 1, &normalTexture);
	deviceContext->PSSetShaderResources(1, 1, &diffuseTexture);
	deviceContext->PSSetShaderResources(2, 1, &specularTexture);
	deviceContext->PSSetShaderResources(3, 1, &positionTexture);

}

void DeferredShader::Render(ID3D11DeviceContext* deviceContext, int indexCount)
{
	// Set the sampler state in the pixel shader.
	deviceContext->PSSetSamplers(0, 1, &m_sampleState);

	// Base render function.
	BaseShader::Render(deviceContext, indexCount);
}
