#include "PrePassLightShaderMSAA.h"
#include "../DXFramework/ResourceContainer.h"

PrePassLightShaderMSAA::PrePassLightShaderMSAA(ID3D11Device* device, HWND hwnd, bool msaaPass) : BaseShader(device, hwnd)
{
	m_deferredShader = true;
	m_msaaPass = msaaPass;
	InitShader(L"../res/shaders/prepass_light_vs.hlsl", L"../res/shaders/prepass_light_msaa_ps.hlsl");
}


PrePassLightShaderMSAA::~PrePassLightShaderMSAA()
{
	// Release the sampler state.
	if (m_sampleState)
	{
		m_sampleState->Release();
		m_sampleState = 0;
	}

	// Release the layout.
	if (m_layout)
	{
		m_layout->Release();
		m_layout = 0;
	}

	//release the matrix buffer
	if (m_matrixBuffer)
	{
		m_matrixBuffer->Release();
		m_matrixBuffer = 0;
	}

	// Release the light constant buffer.
	if (m_lightBuffer)
	{
		m_lightBuffer->Release();
		m_lightBuffer = 0;
	}

	//Release the camera constant buffer
	if (m_cameraBuffer)
	{
		m_cameraBuffer->Release();
		m_cameraBuffer = 0;
	}

	//Release base shader components
	BaseShader::~BaseShader();
}


void PrePassLightShaderMSAA::InitShader(WCHAR* vsFilename, WCHAR* psFilename)
{
	HRESULT result;
	D3D11_SAMPLER_DESC samplerDesc;
	D3D11_BLEND_DESC blendStateDescription;
	D3D11_DEPTH_STENCIL_DESC depthStencilDescription;

	loadVertexShader(vsFilename);

	loadPixelShader(psFilename, (m_msaaPass) ? "ssMain" : "main");
	
	// Create a texture sampler state description.
	samplerDesc.Filter = D3D11_FILTER_MIN_MAG_MIP_LINEAR;
	samplerDesc.AddressU = D3D11_TEXTURE_ADDRESS_BORDER;
	samplerDesc.AddressV = D3D11_TEXTURE_ADDRESS_BORDER;
	samplerDesc.AddressW = D3D11_TEXTURE_ADDRESS_BORDER;
	samplerDesc.MipLODBias = 0.0f;
	samplerDesc.MaxAnisotropy = 16;
	samplerDesc.ComparisonFunc = D3D11_COMPARISON_ALWAYS;
	samplerDesc.BorderColor[0] = 1e5f;
	samplerDesc.BorderColor[1] = 0;
	samplerDesc.BorderColor[2] = 0;
	samplerDesc.BorderColor[3] = 0;
	samplerDesc.MinLOD = 0;
	samplerDesc.MaxLOD = D3D11_FLOAT32_MAX;

	// Create the texture sampler state.
	result = m_device->CreateSamplerState(&samplerDesc, &m_sampleState);
	if (FAILED(result))
	{
		printf("Sample State creation failed.");
	}
	
	// Clear the blend state description.
	ZeroMemory(&blendStateDescription, sizeof(D3D11_BLEND_DESC));

	// Create an alpha disabled blend state description.
	blendStateDescription.RenderTarget[0].BlendEnable = TRUE;
	blendStateDescription.RenderTarget[0].SrcBlend = D3D11_BLEND_ONE;
	blendStateDescription.RenderTarget[0].DestBlend = D3D11_BLEND_ONE;
	blendStateDescription.RenderTarget[0].BlendOp = D3D11_BLEND_OP_ADD;
	blendStateDescription.RenderTarget[0].SrcBlendAlpha = D3D11_BLEND_ONE;
	blendStateDescription.RenderTarget[0].DestBlendAlpha = D3D11_BLEND_ZERO;
	blendStateDescription.RenderTarget[0].BlendOpAlpha = D3D11_BLEND_OP_ADD;
	blendStateDescription.RenderTarget[0].RenderTargetWriteMask = 0x0f;
	blendStateDescription.AlphaToCoverageEnable = false;
	blendStateDescription.IndependentBlendEnable = false;

	// Create the blend state using the description.
	result = m_device->CreateBlendState(&blendStateDescription, &m_blendState);

	// Create the depth disabled view required for this shader
	depthStencilDescription.DepthEnable = false;
	depthStencilDescription.DepthWriteMask = D3D11_DEPTH_WRITE_MASK_ALL;
	depthStencilDescription.DepthFunc = D3D11_COMPARISON_LESS;
	depthStencilDescription.StencilEnable = true;
	depthStencilDescription.StencilReadMask = 0xFF;
	depthStencilDescription.StencilWriteMask = 0xFF;
	depthStencilDescription.FrontFace.StencilFailOp = D3D11_STENCIL_OP_KEEP;
	depthStencilDescription.FrontFace.StencilDepthFailOp = D3D11_STENCIL_OP_INCR;
	depthStencilDescription.FrontFace.StencilPassOp = D3D11_STENCIL_OP_KEEP;
	depthStencilDescription.FrontFace.StencilFunc = D3D11_COMPARISON_LESS_EQUAL;
	depthStencilDescription.BackFace.StencilFailOp = D3D11_STENCIL_OP_KEEP;
	depthStencilDescription.BackFace.StencilDepthFailOp = D3D11_STENCIL_OP_DECR;
	depthStencilDescription.BackFace.StencilPassOp = D3D11_STENCIL_OP_KEEP;
	depthStencilDescription.BackFace.StencilFunc = D3D11_COMPARISON_LESS_EQUAL;

	// Create the state using the device.
	m_device->CreateDepthStencilState(&depthStencilDescription, &m_depthStencilState);

	//Create the light constant buffer
	m_lightBuffer = CreateConstantBuffer(m_device, sizeof(LightBufferType), true, true, NULL);

	//Create the camera constant buffer
	m_cameraBuffer = CreateConstantBuffer(m_device, sizeof(CameraBufferType), true, true, NULL);

	//Create the vertex shader matrix buffer
	m_matrixBuffer = CreateConstantBuffer(m_device, sizeof(MatrixBufferType), true, true, NULL);

}

void PrePassLightShaderMSAA::Render(ID3D11DeviceContext* deviceContext, D3D* d3d, void* resourceContainer)
{
	HRESULT result;
	D3D11_MAPPED_SUBRESOURCE mappedResource;
	LightBufferType* lightPtr;
	CameraBufferType* cameraPtr;
	MatrixBufferType* matrixPtr;
	unsigned int bufferNumber;
	XMMATRIX view, proj;

	ResourceContainer* container = reinterpret_cast<ResourceContainer*>(resourceContainer);

	// Set the light buffer as the render target
	container->GetBuffers(L"LIGHT_BUFFER")->SetRenderTargets(container->GetBuffers(L"G_BUFFER")->GetDepthStencilView());
	container->GetBuffers(L"LIGHT_BUFFER")->ClearRenderTargets(0.0f, 0.0f, 0.0f, 1.0f);

	// Set the sampler state in the pixel shader.
	deviceContext->PSSetSamplers(0, 1, &m_sampleState);

	// Send the mesh data to the GPU
	container->GetMesh(m_renderObjects[0]->GetMeshIndex(0))->SendData(deviceContext);
	m_indexCount = container->GetMesh(m_renderObjects[0]->GetMeshIndex(0))->GetIndexCount();

	// Lock the camera constant buffer so it can be written to
	result = deviceContext->Map(m_cameraBuffer, 0, D3D11_MAP_WRITE_DISCARD, 0, &mappedResource);

	// Get a pointer to the constant buffer data
	cameraPtr = (CameraBufferType*)mappedResource.pData;

	// Copy the camera position into the constant buffer
	cameraPtr->cameraPos = container->GetCamera(m_renderCameras[0])->GetPosition();
	cameraPtr->padding = 0.0f;

	// Unlock the constant buffer
	deviceContext->Unmap(m_cameraBuffer, 0);

	bufferNumber = 1;

	// Set the constant buffer in the pixel shader
	deviceContext->PSSetConstantBuffers(bufferNumber, 1, &m_cameraBuffer);

	// Send the inverse view projection matrix to the pixel shader
	container->GetCamera(m_renderCameras[0])->GetViewMatrix(view);
	d3d->GetProjectionMatrix(proj);

	// Set the g buffer textures in the pixel shader.
	ID3D11ShaderResourceView* normalTexture = container->GetBuffers(L"G_BUFFER")->GetShaderResourceView(0);
	ID3D11ShaderResourceView* positionTexture = container->GetBuffers(L"G_BUFFER")->GetShaderResourceView(1);

	deviceContext->PSSetShaderResources(0, 1, &normalTexture);
	deviceContext->PSSetShaderResources(1, 1, &positionTexture);

	for (auto light : m_renderLights)
	{
		Light* renderLight = container->GetLight(light);

		// Depending on light type either send the shadow map or ambient occlusion texture to the shader
		if (renderLight->GetLightType() == 0)
		{
			ID3D11ShaderResourceView* ambientTexture = container->GetBuffers(L"AMBIENT_OCCLUSION_TEXTURE")->GetShaderResourceView(0);
			deviceContext->PSSetShaderResources(2, 1, &ambientTexture);
		}
		else
		{
			ID3D11ShaderResourceView* shadowMap = renderLight->GetShadowMap()->GetShaderResourceView(0);
			deviceContext->PSSetShaderResources(2, 1, &shadowMap);
		}

		// send matrix data to the pixel shader
		deviceContext->Map(m_matrixBuffer, 0, D3D11_MAP_WRITE_DISCARD, 0, &mappedResource);
		matrixPtr = (MatrixBufferType*)mappedResource.pData;
		matrixPtr->lightView = XMMatrixTranspose(renderLight->GetViewMatrix());
		matrixPtr->lightProj = XMMatrixTranspose(renderLight->GetProjectionMatrix());
		deviceContext->Unmap(m_matrixBuffer, 0);
		bufferNumber = 2;
		deviceContext->PSSetConstantBuffers(bufferNumber, 1, &m_matrixBuffer);

		// Send light data to pixel shader
		deviceContext->Map(m_lightBuffer, 0, D3D11_MAP_WRITE_DISCARD, 0, &mappedResource);
		lightPtr = (LightBufferType*)mappedResource.pData;
		XMFLOAT3 pos = renderLight->GetPosition();
		lightPtr->lightPos = XMFLOAT4(pos.x, pos.y, pos.z, 1.0f);
		lightPtr->lightColor = renderLight->GetDiffuseColour();
		XMFLOAT3 dir = renderLight->GetDirection();
		lightPtr->lightDirection = XMFLOAT4(dir.x, dir.y, dir.z, 1.0f);
		lightPtr->lightRange = XMFLOAT4(25.0f, 25.0f, 25.0f, 25.0f);
		lightPtr->spotlightAngles = XMFLOAT2(45.0f, 45.0f);
		lightPtr->lightType = renderLight->GetLightType();
		lightPtr->padding = container->GetFloat(m_renderFloats[0])->GetParameter(0);

		deviceContext->Unmap(m_lightBuffer, 0);
		bufferNumber = 0;
		deviceContext->PSSetConstantBuffers(bufferNumber, 1, &m_lightBuffer);

		BaseShader::Render(deviceContext, d3d, resourceContainer);			
		
		ID3D11ShaderResourceView* nullSRV[] = { NULL };
		deviceContext->PSSetShaderResources(2, 1, nullSRV);
	}
}