#include "ShadowMapShader.h"
#include "../DXFramework/ResourceContainer.h"

ShadowMapShader::ShadowMapShader(ID3D11Device* device, HWND hwnd) : BaseShader(device, hwnd)
{
	InitShader(L"../res/shaders/shadow_map_vs.hlsl", L"../res/shaders/shadow_map_ps.hlsl");
}

ShadowMapShader::~ShadowMapShader()
{
	if (m_lightDataBuffer)
	{
		m_lightDataBuffer->Release();
		m_lightDataBuffer = 0;
	}

	if (m_matrixBuffer)
	{
		m_matrixBuffer->Release();
		m_matrixBuffer = 0;
	}

	BaseShader::~BaseShader();
}

void ShadowMapShader::InitShader(WCHAR* vsFilename, WCHAR* psFilename)
{
	D3D11_BLEND_DESC blendStateDescription;

	loadVertexShader(vsFilename);
	loadPixelShader(psFilename);

	m_matrixBuffer = CreateConstantBuffer(m_device, sizeof(LightMatrixBuffer), true, true, NULL);
	m_lightDataBuffer = CreateConstantBuffer(m_device, sizeof(LightDataBuffer), true, true, NULL);

	// Clear the blend state description.
	ZeroMemory(&blendStateDescription, sizeof(D3D11_BLEND_DESC));

	// Create an alpha disabled blend state description.
	blendStateDescription.RenderTarget[0].BlendEnable = TRUE;
	blendStateDescription.RenderTarget[0].SrcBlend = D3D11_BLEND_ONE;
	blendStateDescription.RenderTarget[0].DestBlend = D3D11_BLEND_ONE;
	blendStateDescription.RenderTarget[0].BlendOp = D3D11_BLEND_OP_ADD;
	blendStateDescription.RenderTarget[0].SrcBlendAlpha = D3D11_BLEND_ONE;
	blendStateDescription.RenderTarget[0].DestBlendAlpha = D3D11_BLEND_ZERO;
	blendStateDescription.RenderTarget[0].BlendOpAlpha = D3D11_BLEND_OP_ADD;
	blendStateDescription.RenderTarget[0].RenderTargetWriteMask = 0x0f;
	blendStateDescription.AlphaToCoverageEnable = false;
	blendStateDescription.IndependentBlendEnable = false;

	// Create the blend state using the description.
	//m_device->CreateBlendState(&blendStateDescription, &m_blendState);
}

void ShadowMapShader::Render(ID3D11DeviceContext* deviceContext, D3D* d3d, void* resourceContainer)
{
	D3D11_MAPPED_SUBRESOURCE mappedResource;
	LightMatrixBuffer* matrixBuffer;
	LightDataBuffer* lightBuffer;
	XMMATRIX world, view, proj;
	int bufferNumber;
	ResourceContainer* container = reinterpret_cast<ResourceContainer*>(resourceContainer);

	for (auto light : m_renderLights)
	{
		Light* current_light = container->GetLight(light);

		// set the lights shadow map as output of the shader
		current_light->GetShadowMap()->SetRenderTargets();
		current_light->GetShadowMap()->ClearRenderTargets(0.0f, 0.0f, 0.0f, 0.0f);

		// Send the light data to the vertex shader
		deviceContext->Map(m_lightDataBuffer, 0, D3D11_MAP_WRITE_DISCARD, 0, &mappedResource);
		lightBuffer = (LightDataBuffer*)mappedResource.pData;
		lightBuffer->lightPos = current_light->GetPosition();
		lightBuffer->padding = 0.0f;
		deviceContext->Unmap(m_lightDataBuffer, 0);
		bufferNumber = 1;
		deviceContext->VSSetConstantBuffers(bufferNumber, 1, &m_lightDataBuffer);

		view = current_light->GetViewMatrix();
		proj = current_light->GetProjectionMatrix();

		for (auto object : m_renderObjects)
		{
			world = object->GetTransformMatrix();

			// Send the matrix data to the buffer
			deviceContext->Map(m_matrixBuffer, 0, D3D11_MAP_WRITE_DISCARD, 0, &mappedResource);
			matrixBuffer = (LightMatrixBuffer*)mappedResource.pData;
			matrixBuffer->world = XMMatrixTranspose(world);
			matrixBuffer->lightView = XMMatrixTranspose(view);
			matrixBuffer->lightProj = XMMatrixTranspose(proj);
			deviceContext->Unmap(m_matrixBuffer, 0);
			bufferNumber = 0;
			deviceContext->VSSetConstantBuffers(bufferNumber, 1, &m_matrixBuffer);

			for (auto mesh : object->GetMeshes())
			{
				container->GetMesh(mesh)->SendData(deviceContext);
				m_indexCount = container->GetMesh(mesh)->GetIndexCount();
				m_renderType = container->GetMesh(mesh)->GetRenderType();

				BaseShader::Render(deviceContext, d3d, resourceContainer);
			}
		}
	}
}