#include "SkyboxShader.h"
#include "../DXFramework/ResourceContainer.h"

SkyboxShader::SkyboxShader(ID3D11Device * device, HWND hwnd) : BaseShader(device, hwnd)
{
	InitShader(L"../res/shaders/skybox_vs.hlsl", L"../res/shaders/skybox_ps.hlsl");
}

SkyboxShader::~SkyboxShader()
{
	// release the vs matrix buffer
	if (m_vsMatrixBuffer)
	{
		m_vsMatrixBuffer->Release();
		m_vsMatrixBuffer = 0;
	}

	// release the ps gradient buffer
	if (m_psGradientBuffer)
	{
		m_psGradientBuffer->Release();
		m_psGradientBuffer = 0;
	}

	// release the depth stencil state
	if (m_depthStencilState)
	{
		m_depthStencilState->Release();
		m_depthStencilState = 0;
	}

	// release the rasterizer state
	if (m_rasterizerState)
	{
		m_rasterizerState->Release();
		m_rasterizerState = 0;
	}
}

void SkyboxShader::InitShader(WCHAR* vsFilename, WCHAR* psFilename)
{
	HRESULT result;
	
	loadVertexShader(vsFilename);
	loadPixelShader(psFilename);

	// create constant buffers
	m_vsMatrixBuffer = CreateConstantBuffer(m_device, sizeof(VSMatrixBufferType), true, true, NULL);
	m_psGradientBuffer = CreateConstantBuffer(m_device, sizeof(PSGradientBufferType), true, true, NULL);

	// create a backface culling disabled raster state
	D3D11_RASTERIZER_DESC rasterDesc;
	rasterDesc.AntialiasedLineEnable = TRUE;
	rasterDesc.CullMode = D3D11_CULL_NONE;
	rasterDesc.DepthBias = 0;
	rasterDesc.DepthBiasClamp = 0.0f;
	rasterDesc.DepthClipEnable = true;
	rasterDesc.FillMode = D3D11_FILL_SOLID;
	rasterDesc.FrontCounterClockwise = true;
	rasterDesc.MultisampleEnable = true;
	rasterDesc.ScissorEnable = false;
	rasterDesc.SlopeScaledDepthBias = 0.0f;

	m_device->CreateRasterizerState(&rasterDesc, &m_rasterizerState);

	// create a depth write disabled depth stencil state
	D3D11_DEPTH_STENCIL_DESC depthStencilDesc;
	
	// Set up the description of the stencil state.
	depthStencilDesc.DepthEnable = false;
	depthStencilDesc.DepthWriteMask = D3D11_DEPTH_WRITE_MASK_ZERO;
	depthStencilDesc.DepthFunc = D3D11_COMPARISON_LESS;

	depthStencilDesc.StencilEnable = false;
	depthStencilDesc.StencilReadMask = 0xFF;
	depthStencilDesc.StencilWriteMask = 0xFF;

	// Stencil operations if pixel is front-facing.
	depthStencilDesc.FrontFace.StencilFailOp = D3D11_STENCIL_OP_KEEP;
	depthStencilDesc.FrontFace.StencilDepthFailOp = D3D11_STENCIL_OP_INCR;
	depthStencilDesc.FrontFace.StencilPassOp = D3D11_STENCIL_OP_REPLACE;
	depthStencilDesc.FrontFace.StencilFunc = D3D11_COMPARISON_ALWAYS;

	// Stencil operations if pixel is back-facing.
	depthStencilDesc.BackFace.StencilFailOp = D3D11_STENCIL_OP_KEEP;
	depthStencilDesc.BackFace.StencilDepthFailOp = D3D11_STENCIL_OP_DECR;
	depthStencilDesc.BackFace.StencilPassOp = D3D11_STENCIL_OP_REPLACE;
	depthStencilDesc.BackFace.StencilFunc = D3D11_COMPARISON_ALWAYS;

	// Create the depth stencil state.
	m_device->CreateDepthStencilState(&depthStencilDesc, &m_depthStencilState);
}

void SkyboxShader::Render(ID3D11DeviceContext* deviceContext, D3D* d3d, void * resourceContainer)
{
	HRESULT result;
	D3D11_MAPPED_SUBRESOURCE mappedResource;
	VSMatrixBufferType* vsMatrixBuffer;
	PSGradientBufferType* psGradientBuffer;
	unsigned int bufferNumber;
	XMMATRIX world, view, proj;

	ResourceContainer* container = reinterpret_cast<ResourceContainer*>(resourceContainer);

	container->GetCamera(L"MAIN_CAMERA")->GetViewMatrix(view);
	d3d->GetProjectionMatrix(proj);
	XMFLOAT3 cameraPos = container->GetCamera(L"MAIN_CAMERA")->GetPosition();
	
	// send skybox mesh data to the GPU
	BaseMesh* skySphereMesh = container->GetMesh(L"SKY_SPHERE_MESH");
	skySphereMesh->SendData(deviceContext);
	m_indexCount = skySphereMesh->GetIndexCount();

	// send the matrix to the vertex shader
	deviceContext->Map(m_vsMatrixBuffer, 0, D3D11_MAP_WRITE_DISCARD, 0, &mappedResource);
	vsMatrixBuffer = (VSMatrixBufferType*)mappedResource.pData;
	vsMatrixBuffer->world = XMMatrixTranspose(XMMatrixTranslation(cameraPos.x, cameraPos.y, cameraPos.z));
	vsMatrixBuffer->view = XMMatrixTranspose(view);
	vsMatrixBuffer->proj = XMMatrixTranspose(proj);
	deviceContext->Unmap(m_vsMatrixBuffer, 0);
	bufferNumber = 0;
	deviceContext->VSSetConstantBuffers(bufferNumber, 1, &m_vsMatrixBuffer);

	// send the gradient data to the pixel shader
	deviceContext->Map(m_psGradientBuffer, 0, D3D11_MAP_WRITE_DISCARD, 0, &mappedResource);
	psGradientBuffer = (PSGradientBufferType*)mappedResource.pData;
	psGradientBuffer->apexColor = { 0.0f, 0.15f, 0.66f, 1.0f };
	psGradientBuffer->centerColor = { 0.81f, 0.38f, 0.66f, 1.0f };
	deviceContext->Unmap(m_psGradientBuffer, 0);
	bufferNumber = 0;
	deviceContext->PSSetConstantBuffers(bufferNumber, 1, &m_psGradientBuffer);

	// set the buffer output
	d3d->SetBackBufferRenderTarget();
	d3d->ResetViewport();

	BaseShader::Render(deviceContext, d3d, resourceContainer);

	// reset the raster and depth stencil states
	deviceContext->RSSetState(d3d->GetDefaultRasterizerState());
	deviceContext->OMSetDepthStencilState(d3d->GetDefaultDepthStencilState(), 1);
}

