#include "TerrainMesh.h"

TerrainMesh::TerrainMesh()
{

}

TerrainMesh::TerrainMesh(ID3D11Device* device, ID3D11DeviceContext* deviceContext, WCHAR* textureFilename, int x_len, int z_len)
{
	m_terrainXLength = x_len;
	m_terrainZLength = z_len;

	InitBuffers(device);
}

TerrainMesh::~TerrainMesh()
{
}

void TerrainMesh::InitBuffers(ID3D11Device* device)
{
	HRESULT result;
	VertexType* vertices;	
	D3D11_BUFFER_DESC vertexBufferDesc, indexBufferDesc;
	D3D11_SUBRESOURCE_DATA vertexData, indexData;
	unsigned long* indices;
	int vertex, index;

	m_vertexCount = (m_terrainXLength + 1) * (m_terrainZLength + 1);
	m_indexCount = (m_terrainXLength * m_terrainZLength) * 12;

	vertices = new VertexType[m_vertexCount];
	indices = new unsigned long[m_indexCount];

	float fWidth = (float)m_terrainXLength;
	float fHeight = (float)m_terrainZLength;

	for (int x = 0; x < m_terrainXLength + 1; ++x)
	{
		for (int z = 0; z < m_terrainZLength + 1; ++z)
		{
			float fX = ((float)x / fWidth) - 0.5f;
			float fZ = ((float)z / fHeight) - 0.5f;
			vertex = x + z * (m_terrainXLength + 1);
			vertices[vertex].position = XMFLOAT3(fX, 0.0f, fZ);
			vertices[vertex].texture = XMFLOAT2(fX + 0.5f, fZ + 0.5f);
		}
	}

	index = 0;

	for (int x = 0; x < m_terrainXLength; ++x)
	{
		for (int z = 0; z < m_terrainZLength; ++z)
		{
			// Define 12 control points per terrain quad
			// Indices 0-3 are the vertices of the quad itself
			indices[index] = (z + 0) + (x + 0) * (m_terrainXLength + 1);
			index++;
			indices[index] = (z + 1) + (x + 0) * (m_terrainXLength + 1);
			index++;
			indices[index] = (z + 0) + (x + 1) * (m_terrainXLength + 1);
			index++;
			indices[index] = (z + 1) + (x + 1) * (m_terrainXLength + 1);
			index++;

			// 4-5 are +x
			indices[index] = clamp(z + 0, 0, m_terrainZLength) + clamp(x + 2, 0, m_terrainXLength) * (m_terrainXLength + 1);
			index++;
			indices[index] = clamp(z + 1, 0, m_terrainZLength) + clamp(x + 2, 0, m_terrainXLength) * (m_terrainXLength + 1);
			index++;
			// 6-7 are +z
			indices[index] = clamp(z + 2, 0, m_terrainZLength) + clamp(x + 0, 0, m_terrainXLength) * (m_terrainXLength + 1);
			index++;
			indices[index] = clamp(z + 2, 0, m_terrainZLength) + clamp(x + 1, 0, m_terrainXLength) * (m_terrainXLength + 1);
			index++;
			
			// 8-9 are -x
			indices[index] = clamp(z + 0, 0, m_terrainZLength) + clamp(x - 1, 0, m_terrainXLength) * (m_terrainXLength + 1);
			index++;
			indices[index] = clamp(z + 1, 0, m_terrainZLength) + clamp(x - 1, 0, m_terrainXLength) * (m_terrainXLength + 1);
			index++;
			// 10-11 are -z
			indices[index] = clamp(z - 1, 0, m_terrainZLength) + clamp(x + 0, 0, m_terrainXLength) * (m_terrainXLength + 1);
			index++;
			indices[index] = clamp(z - 1, 0, m_terrainZLength) + clamp(x + 1, 0, m_terrainXLength) * (m_terrainXLength + 1);
			index++;
		}
	}

	// Set up the description of the static vertex buffer.
	vertexBufferDesc.Usage = D3D11_USAGE_DEFAULT;
	vertexBufferDesc.ByteWidth = sizeof(VertexType)* m_vertexCount;
	vertexBufferDesc.BindFlags = D3D11_BIND_VERTEX_BUFFER;
	vertexBufferDesc.CPUAccessFlags = 0;
	vertexBufferDesc.MiscFlags = 0;
	vertexBufferDesc.StructureByteStride = 0;

	// Give the subresource structure a pointer to the vertex data.
	vertexData.pSysMem = vertices;
	vertexData.SysMemPitch = 0;
	vertexData.SysMemSlicePitch = 0;

	// Now create the vertex buffer.
	result = device->CreateBuffer(&vertexBufferDesc, &vertexData, &m_vertexBuffer);

	// Set up the description of the static index buffer.
	indexBufferDesc.Usage = D3D11_USAGE_DEFAULT;
	indexBufferDesc.ByteWidth = sizeof(unsigned long)* m_indexCount;
	indexBufferDesc.BindFlags = D3D11_BIND_INDEX_BUFFER;
	indexBufferDesc.CPUAccessFlags = 0;
	indexBufferDesc.MiscFlags = 0;
	indexBufferDesc.StructureByteStride = 0;

	// Give the subresource structure a pointer to the index data.
	indexData.pSysMem = indices;
	indexData.SysMemPitch = 0;
	indexData.SysMemSlicePitch = 0;

	// Create the index buffer.
	result = device->CreateBuffer(&indexBufferDesc, &indexData, &m_indexBuffer);

	// Release the arrays now that the buffers have been created and loaded.
	delete[] vertices;
	vertices = 0;

	delete[] indices;
	indices = 0;
}

void TerrainMesh::SendData(ID3D11DeviceContext* deviceContext)
{
	unsigned int stride;
	unsigned int offset;

	// Set vertex buffer stride and offset.
	stride = (m_boneAnimated) ? sizeof(AnimVertexType) : sizeof(VertexType);
	offset = 0;

	// Set the vertex buffer to active in the input assembler so it can be rendered.
	deviceContext->IASetVertexBuffers(0, 1, &m_vertexBuffer, &stride, &offset);

	// Set the index buffer to active in the input assembler so it can be rendered.
	deviceContext->IASetIndexBuffer(m_indexBuffer, DXGI_FORMAT_R32_UINT, 0);

	// Set the type of primitive that should be rendered from this vertex buffer, in this case triangles.
	deviceContext->IASetPrimitiveTopology(D3D11_PRIMITIVE_TOPOLOGY_12_CONTROL_POINT_PATCHLIST);
}