#include "TerrainShader.h"
#include "../DXFramework/ResourceContainer.h"

TerrainShader::TerrainShader(ID3D11Device* device, HWND hwnd) : BaseShader(device, hwnd)
{
	m_deferredShader = false;
	InitShader(L"../res/shaders/tessellated_terrain_vs.hlsl", L"../res/shaders/tessellated_terrain_hs.hlsl", L"../res/shaders/tessellated_terrain_ds.hlsl", L"../res/shaders/tessellated_terrain_gs.hlsl", L"../res/shaders/tessellated_terrain_ps.hlsl");
}

TerrainShader::~TerrainShader()
{
	// Release the sampler
	if (m_sampleState)
	{
		m_sampleState->Release();
		m_sampleState = 0;
	}

	// Release the vs matrix buffer
	if (m_vsMatrixBuffer)
	{
		m_vsMatrixBuffer->Release();
		m_vsMatrixBuffer = 0;
	}

	// Release the vs noise buffer
	if (m_vsNoiseBuffer)
	{
		m_vsNoiseBuffer->Release();
		m_vsNoiseBuffer = 0;
	}

	// Release the hs camera buffer
	if (m_hsCameraBuffer)
	{
		m_hsCameraBuffer->Release();
		m_hsCameraBuffer = 0;
	}

	// Release the hs param buffer
	if (m_hsLODParamBuffer)
	{
		m_hsLODParamBuffer->Release();
		m_hsLODParamBuffer = 0;
	}

	// Release the ds matrix buffer
	if (m_matrixBuffer)
	{
		m_matrixBuffer->Release();
		m_matrixBuffer = 0;
	}

	// Release the ds param buffer
	if (m_dsSampleParamBuffer)
	{
		m_dsSampleParamBuffer->Release();
		m_dsSampleParamBuffer = 0;
	}

	
}

void TerrainShader::InitShader(WCHAR* vsFilename, WCHAR* hsFilename, WCHAR* dsFilename, WCHAR* gsFilename, WCHAR* psFilename)
{

	InitShader(vsFilename, psFilename);

	loadHullShader(hsFilename);
	loadDomainShader(dsFilename);

	D3D11_SO_DECLARATION_ENTRY soDecl[] = 
	{
		{0, "POSITION", 0, 0, 3, 0},
		{0, "TEXCOORD", 0, 0, 2, 0},
		{0, "NORMAL", 0, 0, 3, 0},
		{0, "TANGENT", 0, 0, 4, 0}
	};


	loadGeometryShaderWithStreamOutput(gsFilename, soDecl, sizeof(soDecl)/ sizeof(soDecl[0]), 1);
}

void TerrainShader::InitShader(WCHAR* vsFilename, WCHAR* psFilename)
{
	HRESULT result;
	D3D11_SAMPLER_DESC samplerDesc;


	loadVertexShader(vsFilename);

	loadPixelShader(psFilename);

	// Create all the necessary constant buffers for this render effect
	m_vsMatrixBuffer = CreateConstantBuffer(m_device, sizeof(VSMatrixBufferType), true, true, NULL);
	m_vsNoiseBuffer = CreateConstantBuffer(m_device, sizeof(VSNoiseBufferType), true, true, NULL);
	m_hsCameraBuffer = CreateConstantBuffer(m_device, sizeof(HSCameraBufferType), true, true, NULL);
	m_hsLODParamBuffer = CreateConstantBuffer(m_device, sizeof(HSLODParamBufferType), true, true, NULL);
	m_hsSampleParamBuffer = CreateConstantBuffer(m_device, sizeof(DSSampleParamBufferType), true, true, NULL);
	m_matrixBuffer = CreateConstantBuffer(m_device, sizeof(DSMatrixBufferType), true, true, NULL);
	m_dsSampleParamBuffer = CreateConstantBuffer(m_device, sizeof(DSSampleParamBufferType), true, true, NULL);
	m_gsMatrixBuffer = CreateConstantBuffer(m_device, sizeof(GSMatrixBufferType), true, true, NULL);
	m_psLightBuffer = CreateConstantBuffer(m_device, sizeof(LightBufferType), true, true, NULL);
	m_psMaterialBuffer = CreateConstantBuffer(m_device, sizeof(PSMaterialBufferType), true, true, NULL);
	m_psCameraBuffer = CreateConstantBuffer(m_device, sizeof(PSCameraBufferType), true, true, NULL);

	// Create a texture sampler state description.
	samplerDesc.Filter = D3D11_FILTER_MIN_MAG_MIP_LINEAR;
	samplerDesc.AddressU = D3D11_TEXTURE_ADDRESS_CLAMP;
	samplerDesc.AddressV = D3D11_TEXTURE_ADDRESS_CLAMP;
	samplerDesc.AddressW = D3D11_TEXTURE_ADDRESS_CLAMP;
	samplerDesc.MipLODBias = 0.0f;
	samplerDesc.MaxAnisotropy = 16;
	samplerDesc.ComparisonFunc = D3D11_COMPARISON_ALWAYS;
	samplerDesc.BorderColor[0] = 0;
	samplerDesc.BorderColor[1] = 0;
	samplerDesc.BorderColor[2] = 0;
	samplerDesc.BorderColor[3] = 0;
	samplerDesc.MinLOD = 0;
	samplerDesc.MaxLOD = D3D11_FLOAT32_MAX;

	// Create the texture sampler state.
	result = m_device->CreateSamplerState(&samplerDesc, &m_sampleState);
	if (FAILED(result))
	{
		printf("Sample State creation failed.");
	}
}

void TerrainShader::Render(ID3D11DeviceContext* deviceContext, D3D* d3d, void* resourceContainer)
{
	HRESULT result;
	D3D11_MAPPED_SUBRESOURCE mappedResource;
	VSMatrixBufferType* vsMatrixBuffer;
	VSNoiseBufferType* vsNoiseBuffer;
	HSCameraBufferType* hsCameraBuffer;
	HSLODParamBufferType* hsLODParamBuffer;
	DSSampleParamBufferType* hsSampleParamBuffer;
	DSMatrixBufferType* dsMatrixBuffer;
	DSSampleParamBufferType* dsSampleParamBuffer;
	GSMatrixBufferType* gsMatrixBuffer;
	LightBufferType* psLightBuffer;
	PSMaterialBufferType* psMaterialBuffer;
	PSCameraBufferType* psCameraBuffer;

	unsigned int bufferNumber;
	XMMATRIX world, view, proj;

	ResourceContainer* container = reinterpret_cast<ResourceContainer*>(resourceContainer);

	container->GetCamera(L"MAIN_CAMERA")->GetViewMatrix(view);
	d3d->GetProjectionMatrix(proj);

	// Set the sampler state in the pixel shader.
	deviceContext->DSSetSamplers(0, 1, &m_sampleState);

	// Send the mesh data to the GPU - move to for loop if there are multiple terrain meshes for some reason
	container->GetMesh(m_renderObjects[0]->GetMeshIndex(0))->SendData(deviceContext);
	m_indexCount = container->GetMesh(m_renderObjects[0]->GetMeshIndex(0))->GetIndexCount();

	// Send the terrain world matrix to the vertex shader
	deviceContext->Map(m_vsMatrixBuffer, 0, D3D11_MAP_WRITE_DISCARD, 0, &mappedResource);
	vsMatrixBuffer = (VSMatrixBufferType*)mappedResource.pData;
	world = XMMatrixTranspose(m_renderObjects[0]->GetTransformMatrix());
	vsMatrixBuffer->world = world;
	deviceContext->Unmap(m_vsMatrixBuffer, 0);
	bufferNumber = 0;
	deviceContext->VSSetConstantBuffers(bufferNumber, 1, &m_vsMatrixBuffer);

	// Send the noise data to the vertex shader
	deviceContext->Map(m_vsNoiseBuffer, 0, D3D11_MAP_WRITE_DISCARD, 0, &mappedResource);
	vsNoiseBuffer = (VSNoiseBufferType*)mappedResource.pData;
	vsNoiseBuffer->cameraPos = container->GetCamera(m_renderCameras[0])->GetPosition();
	vsNoiseBuffer->terrainSize = 32.0f;
	deviceContext->Unmap(m_vsNoiseBuffer, 0);
	bufferNumber = 1;
	deviceContext->VSSetConstantBuffers(bufferNumber, 1, &m_vsNoiseBuffer);

	// Send the lod lookup uav to the hull shader
	ID3D11ShaderResourceView* lodlookup = container->GetUAV(m_renderUAVs[0])->GetSRV();
	deviceContext->HSSetShaderResources(0, 1, &lodlookup);

	// Send the camera data to the hull shader
	deviceContext->Map(m_hsCameraBuffer, 0, D3D11_MAP_WRITE_DISCARD, 0, &mappedResource);
	hsCameraBuffer = (HSCameraBufferType*)mappedResource.pData;
	hsCameraBuffer->cameraPos = container->GetCamera(m_renderCameras[0])->GetPosition();
	hsCameraBuffer->padding = 1.0f;
	deviceContext->Unmap(m_hsCameraBuffer, 0);
	bufferNumber = 0;
	deviceContext->HSSetConstantBuffers(bufferNumber, 1, &m_hsCameraBuffer);

	// Send the LOD parameters to the hull shader
	deviceContext->Map(m_hsLODParamBuffer, 0, D3D11_MAP_WRITE_DISCARD, 0, &mappedResource);
	hsLODParamBuffer = (HSLODParamBufferType*)mappedResource.pData;
	hsLODParamBuffer->minMaxDistance = XMFLOAT4(4.0f, 30.0f, 0.0f, 0.0f);
	hsLODParamBuffer->minMaxLOD = XMFLOAT4(1.0f, 10.0f, 0.0f, 0.0f);
	deviceContext->Unmap(m_hsLODParamBuffer, 0);
	bufferNumber = 1;
	deviceContext->HSSetConstantBuffers(bufferNumber, 1, &m_hsLODParamBuffer);

	// Send the sampler parameters to the hull shader
	result = deviceContext->Map(m_hsSampleParamBuffer, 0, D3D11_MAP_WRITE_DISCARD, 0, &mappedResource);
	hsSampleParamBuffer = (DSSampleParamBufferType*)mappedResource.pData;
	hsSampleParamBuffer->heightMapDimensions = XMFLOAT4(512, 512, 32, 32);
	deviceContext->Unmap(m_hsSampleParamBuffer, 0);
	bufferNumber = 2;
	deviceContext->HSSetConstantBuffers(bufferNumber, 1, &m_hsSampleParamBuffer);

	// Send the matrix parameters to the domain shader
	deviceContext->Map(m_matrixBuffer, 0, D3D11_MAP_WRITE_DISCARD, 0, &mappedResource);
	dsMatrixBuffer = (DSMatrixBufferType*)mappedResource.pData;
	dsMatrixBuffer->invTposeWorld = XMMatrixTranspose(XMMatrixInverse(NULL, world));
	deviceContext->Unmap(m_matrixBuffer, 0);
	bufferNumber = 0;
	deviceContext->DSSetConstantBuffers(bufferNumber, 1, &m_matrixBuffer);

	// Send the sampler parameters to the domain shader
	result = deviceContext->Map(m_dsSampleParamBuffer, 0, D3D11_MAP_WRITE_DISCARD, 0, &mappedResource);
	dsSampleParamBuffer = (DSSampleParamBufferType*)mappedResource.pData;
	dsSampleParamBuffer->heightMapDimensions = XMFLOAT4(512, 512, 32, 32);
	deviceContext->Unmap(m_dsSampleParamBuffer, 0);
	bufferNumber = 1;
	deviceContext->DSSetConstantBuffers(bufferNumber, 1, &m_dsSampleParamBuffer);

	// Send the heightmap texture to the domain shader
	ID3D11ShaderResourceView* heightmap = container->GetUAV(m_renderUAVs[1])->GetSRV();
	deviceContext->DSSetShaderResources(0, 1, &heightmap);

	// Send the matrix parameters to the geometry shader
	result = deviceContext->Map(m_gsMatrixBuffer, 0, D3D11_MAP_WRITE_DISCARD, 0, &mappedResource);
	gsMatrixBuffer = (GSMatrixBufferType*)mappedResource.pData;
	gsMatrixBuffer->view = XMMatrixTranspose(view);
	gsMatrixBuffer->proj = XMMatrixTranspose(proj);
	deviceContext->Unmap(m_gsMatrixBuffer, 0);
	bufferNumber = 0;
	deviceContext->GSSetConstantBuffers(bufferNumber, 1, &m_gsMatrixBuffer);

	// Set the stream output buffer as the output of the geometry shader
	ID3D11Buffer* outputBuffer = container->GetMesh(m_renderObjects[1]->GetMeshIndex(0))->GetVertexBuffer();
	UINT offset[4] = { 0, 0, 0, 0 };
	ID3D11Buffer* buffers[] =
	{
		outputBuffer,
		NULL,
		NULL,
		NULL
	};

	deviceContext->SOSetTargets(4, buffers, offset);
	
	/*
	// Send the light data to the pixel shader
	Light* light = container->GetLight(0);
	result = deviceContext->Map(m_psLightBuffer, 0, D3D11_MAP_WRITE_DISCARD, 0, &mappedResource);
	psLightBuffer = (LightBufferType*)mappedResource.pData;
	XMFLOAT3 pos = light->GetPosition();
	psLightBuffer->lightPos = XMFLOAT4(pos.x, pos.y, pos.z, 1.0f);
	psLightBuffer->lightColor = light->GetDiffuseColour();
	XMFLOAT3 dir = light->GetDirection();
	psLightBuffer->lightDirection = XMFLOAT4(dir.x, dir.y, dir.z, 1.0f);
	psLightBuffer->lightRange = XMFLOAT4(25.0f, 25.0f, 25.0f, 25.0f);
	psLightBuffer->spotlightAngles = XMFLOAT2(45.0f, 45.0f);
	psLightBuffer->lightType = light->GetLightType();
	psLightBuffer->padding = 0.0f;

	deviceContext->Unmap(m_psLightBuffer, 0);
	bufferNumber = 0;
	deviceContext->PSSetConstantBuffers(bufferNumber, 1, &m_psLightBuffer);

	// Send the material data to the pixel shader
	result = deviceContext->Map(m_psMaterialBuffer, 0, D3D11_MAP_WRITE_DISCARD, 0, &mappedResource);
	psMaterialBuffer = (PSMaterialBufferType*)mappedResource.pData;
	psMaterialBuffer->specularAlbedo = m_renderObjects[0]->GetSpecularAlbedo();
	psMaterialBuffer->specularPower = m_renderObjects[0]->GetSpecularPower();
	deviceContext->Unmap(m_psMaterialBuffer, 0);
	bufferNumber = 1;
	deviceContext->PSSetConstantBuffers(bufferNumber, 1, &m_psMaterialBuffer);

	// Send the camera data to the pixel shader
	result = deviceContext->Map(m_psCameraBuffer, 0, D3D11_MAP_WRITE_DISCARD, 0, &mappedResource);
	psCameraBuffer = (PSCameraBufferType*)mappedResource.pData;
	psCameraBuffer->cameraPos = container->GetCamera(m_renderCameras[0])->GetPosition();
	psCameraBuffer->padding = 1.0f;
	deviceContext->Unmap(m_psCameraBuffer, 0);
	bufferNumber = 2;
	deviceContext->PSSetConstantBuffers(bufferNumber, 1, &m_psCameraBuffer);
	*/
	m_pixelShader = NULL;

	BaseShader::Render(deviceContext, d3d, resourceContainer);

	ID3D11Buffer* nullBuffer = NULL;
	deviceContext->SOSetTargets(1, &nullBuffer, offset);
}