#include "TestScene.h"
#include "../res/shader_settings.hlsli"
#include "SkyboxShader.h"
#include "HeightmapGenerationShader.h"

TestScene::TestScene(D3D* d3d, HWND hwnd, int screenWidth, int screenHeight) : BaseScene(d3d, hwnd, screenWidth, screenHeight)
{
	BaseShader* shader;
	BaseMesh* mesh;
	DeferredBuffers* buffers;
	Camera* camera;
	Light* light;
	UnorderedAccessView* uav;

	m_runTime = 0.0f;
	m_frameCount = 0.0f;

	// Create the main camera
	camera = new Camera();
	camera->SetPosition(0.0f, 0.0f, -10.0f);
	m_resourceContainer->AddCamera(L"MAIN_CAMERA", camera);

	// Create the sphere mesh and brick texture
	mesh = new SphereMesh(m_device, m_deviceContext, NULL);
	Texture* brick_texture = new Texture(m_device, m_deviceContext, L"../res/mountain.dds");
	Texture* bump_texture = new Texture(m_device, m_deviceContext, L"../res/mountain_bump.dds");
	m_resourceContainer->AddTexture(L"stone01.dds", brick_texture);
	m_resourceContainer->AddTexture(L"bump01.dds", bump_texture);

	// Create the g-buffer buffers
	buffers = new DeferredBuffers(m_device, m_deviceContext, screenWidth, screenHeight, 2, SAMPLE_COUNT);

	// Create the stream output mesh
	mesh = 0;
	mesh = new StreamOutputMesh(m_device, m_deviceContext, ((32 * 32) * 1024));

	// Create the tessellated terrain object
	SceneObject* tessellated_terrain = new SceneObject;
	tessellated_terrain->AddMesh(m_resourceContainer->AddMesh(L"TESSELLATED_TERRAIN_MESH", mesh));
	tessellated_terrain->AddTexture(m_resourceContainer->GetTextureIndex(L"stone01.dds"));
	tessellated_terrain->AddTexture(m_resourceContainer->GetTextureIndex(L"bump01.dds"));
	tessellated_terrain->SetName(L"TESSELLATED_TERRAIN");
	tessellated_terrain->SetSpecularAlbedo(0.1f, 0.1, 0.1f);
	tessellated_terrain->SetSpecularPower(1.0f);
	m_objects.push_back(tessellated_terrain);

	// Create the sky sphere
	mesh = new SphereMesh(m_device, m_deviceContext, NULL);
	m_resourceContainer->AddMesh(L"SKY_SPHERE_MESH", mesh);

	// Initialze the water simulation
	InitWaterSimulation();

	// Create the skybox shader and add it to the resource manager
	shader = 0;
	shader = new SkyboxShader(m_device, hwnd);
	m_resourceContainer->AddShader(L"SKY_SPHERE_SHADER", shader);

	// Create the G-Buffer shader and add it to the resource manager
	shader = 0;
	shader = new PrePassGBufferShader(m_device, hwnd);
	shader->AddObject(tessellated_terrain);
	shader->AddBuffers(m_resourceContainer->AddBuffer(L"G_BUFFER", buffers));
	shader->AddCamera(m_resourceContainer->GetCameraIndex(L"MAIN_CAMERA"));
	m_resourceContainer->AddShader(L"G_BUFFER_SHADER", shader);

	// Create the render window mesh
	mesh = 0;
	mesh = new QuadMesh(m_device, m_deviceContext, NULL);

	// Create the render window object
	SceneObject* render_window = new SceneObject;
	render_window->AddMesh(m_resourceContainer->AddMesh(L"RENDER_WINDOW", mesh));
	render_window->SetName(L"RENDER_WINDOW");
	m_objects.push_back(render_window);

	// Create the Mask shader and add it to the resource manager
	shader = 0;
	shader = new PrePassMaskShader(m_device, hwnd);
	shader->AddObject(render_window);
	shader->AddBuffers(m_resourceContainer->GetBufferIndex(L"G_BUFFER"));
	m_resourceContainer->AddShader(L"MASK_SHADER", shader);

	// Create the light buffer
	buffers = 0;
	buffers = new DeferredBuffers(m_device, m_deviceContext, screenWidth, screenHeight, 1, SAMPLE_COUNT);
	
	// Create an ambient light
	light = 0;
	light = new Light;
	light->SetDirection(0.0f, 0.0f, 0.0f);
	light->SetDiffuseColour(0.4f, 0.4f, 0.4f, 1.0f);
	light->SetPosition(0.0f, 0.0f, 0.0f);
	light->SetConstantAttenuation(1.0f);
	light->SetLinearAttenuation(0.125f);
	light->SetQuadraticAttenuation(0.0f);
	light->SetLightType(0);
	m_resourceContainer->AddLight(L"AMBIENT_LIGHT", light);

	// Create a simple directional light
	light = 0;
	light = new Light;
	light->SetDirection(1.0f, 0.0f, 0.0f);
	light->SetDiffuseColour(1.0f, 1.0f, 1.0f, 1.0f);
	light->SetPosition(-16.0f, 5.0f, 0.0f);
	light->SetConstantAttenuation(1.0f);
	light->SetLinearAttenuation(0.125f);
	light->SetQuadraticAttenuation(0.0f);
	light->SetLightType(1);
	light->GenerateShadowMapBuffer(m_device, m_deviceContext, 1024, 1024, 1, 1);
	light->SetLookAt(1.0f, 0.0f, 0.0f);
	light->GenerateViewMatrix();
	light->GenerateProjectionMatrix(m_Direct3D->GetScreenNear(), m_Direct3D->GetScreenFar()); 
	m_resourceContainer->AddLight(L"DIRECTIONAL_LIGHT", light);

	// Create a second directional light
	light = 0;
	light = new Light;
	light->SetDirection(-1.0f, 0.0f, 0.0f);
	light->SetDiffuseColour(0.8f, 0.0f, 0.0f, 1.0f);
	light->SetPosition(16.0f, 5.0f, 0.0f);
	light->SetConstantAttenuation(1.0f);
	light->SetLinearAttenuation(0.125f);
	light->SetQuadraticAttenuation(0.0f);
	light->SetLightType(1);
	light->GenerateShadowMapBuffer(m_device, m_deviceContext, 1024, 1024, 1, 1);
	light->SetLookAt(-1.0f, 0.0f, 0.0f);
	light->GenerateViewMatrix();
	light->GenerateProjectionMatrix(m_Direct3D->GetScreenNear(), m_Direct3D->GetScreenFar());
	m_resourceContainer->AddLight(L"DIRECTIONAL_LIGHT_02", light);

	// Create the variable that is used to turn shadows on and off
	FloatParameter* floatParam = new FloatParameter(1, NULL);
	floatParam->UpdateParameter(0, 0.0f);

	// Create the Light shader and add it to the resource manager
	shader = 0;
	shader = new PrePassLightShaderMSAA(m_device, hwnd, true);
	shader->AddObject(render_window);
	shader->AddBuffers(m_resourceContainer->AddBuffer(L"LIGHT_BUFFER", buffers));
	shader->AddBuffers(m_resourceContainer->GetBufferIndex(L"G_BUFFER"));
	shader->AddCamera(m_resourceContainer->GetCameraIndex(L"MAIN_CAMERA"));
	shader->AddLight(m_resourceContainer->GetLightIndex(L"AMBIENT_LIGHT"));
	shader->AddLight(m_resourceContainer->GetLightIndex(L"DIRECTIONAL_LIGHT"));
	shader->AddLight(m_resourceContainer->GetLightIndex(L"DIRECTIONAL_LIGHT_02"));
	shader->AddFloat(m_resourceContainer->AddFloat(L"SHADOW_SWITCH", floatParam));
	m_resourceContainer->AddShader(L"LIGHT_SHADER", shader);

	// Create the Final shader and add it to the resource manager
	shader = 0;
	shader = new PrePassFinalShader(m_device, hwnd, true);
	shader->AddObject(tessellated_terrain);
	shader->AddBuffers(m_resourceContainer->GetBufferIndex(L"LIGHT_BUFFER"));
	shader->AddCamera(m_resourceContainer->GetCameraIndex(L"MAIN_CAMERA"));
	m_resourceContainer->AddShader(L"FINAL_SHADER", shader);

	// Initialize the terrain generation
	InitTerrainGeneration();

	// Create the depth storage buffer
	buffers = 0;
	buffers = new DeferredBuffers(m_device, m_deviceContext, screenWidth, screenHeight, 1, 1);

	// Create the depth shader
	shader = 0;
	shader = new DepthShader(m_device, m_hwnd);
	shader->AddCamera(m_resourceContainer->GetCameraIndex(L"MAIN_CAMERA"));
	shader->AddObject(tessellated_terrain);
	shader->AddBuffers(m_resourceContainer->AddBuffer(L"DEPTH_BUFFER", buffers));
	m_resourceContainer->AddShader(L"DEPTH_SHADER", shader);

	// Create the shadow map shader
	shader = 0;
	shader = new ShadowMapShader(m_device, m_hwnd);
	shader->AddLight(m_resourceContainer->GetLightIndex(L"DIRECTIONAL_LIGHT"));
	shader->AddLight(m_resourceContainer->GetLightIndex(L"DIRECTIONAL_LIGHT_02"));
	shader->AddObject(GetObjectByName(L"TESSELLATED_TERRAIN"));
	m_resourceContainer->AddShader(L"SHADOW_MAP_SHADER", shader);

	// Initialize the ambient occlusion shader
	InitAmbientOcclusion();
}

TestScene::~TestScene()
{
}

void TestScene::Update(Input* input, float frameTime)
{
	HandleInput(input, frameTime);

	m_runTime += frameTime;
	m_frameCount++;

	m_resourceContainer->GetCamera(L"MAIN_CAMERA")->Update();

	// Update the required parameters for the water simulation
	float timeFactors[] =
	{
		frameTime,
		1.0f / frameTime,
		m_runTime,
		m_frameCount
	};

	m_resourceContainer->GetFloat(L"WATER_SIM_TIME_FACTORS")->UpdateAllData(timeFactors);
	m_resourceContainer->GetFloat(L"WATER_SIM_TIME_STEP")->UpdateParameter(0, timeFactors[0]);
	
	Render();
}

void TestScene::Render()
{
	// Clear the scene
	m_Direct3D->BeginScene(0.0f, 0.0f, 0.0f, 0.0f);

	// Build the terrain
	m_resourceContainer->GetShader(L"TERRAIN_SHADER")->Render(m_deviceContext, m_Direct3D, m_resourceContainer);

	// Render the depth buffer
	m_resourceContainer->GetShader(L"DEPTH_SHADER")->Render(m_deviceContext, m_Direct3D, m_resourceContainer);

	// Render the ao buffer
	m_resourceContainer->GetShader(L"AMBIENT_OCCLUSION_SHADER")->Render(m_deviceContext, m_Direct3D, m_resourceContainer);

	// Apply a bilateral blur to the ao texture
	m_resourceContainer->GetShader(L"BILATERAL_BLUR_SHADER")->Render(m_deviceContext, m_Direct3D, m_resourceContainer);

	// Generate shadow maps for lights
	m_resourceContainer->GetShader(L"SHADOW_MAP_SHADER")->Render(m_deviceContext, m_Direct3D, m_resourceContainer);

	// Render the skybox into the back buffer
	m_resourceContainer->GetShader(L"SKY_SPHERE_SHADER")->Render(m_deviceContext, m_Direct3D, m_resourceContainer);

	// Execute the deferred renderer
	m_resourceContainer->GetShader(L"G_BUFFER_SHADER")->Render(m_deviceContext, m_Direct3D, m_resourceContainer);
	m_resourceContainer->GetShader(L"MASK_SHADER")->Render(m_deviceContext, m_Direct3D, m_resourceContainer);
	m_resourceContainer->GetShader(L"LIGHT_SHADER")->Render(m_deviceContext, m_Direct3D, m_resourceContainer);	
	m_resourceContainer->GetShader(L"FINAL_SHADER")->Render(m_deviceContext, m_Direct3D, m_resourceContainer);

	// Calculate and render the water simulation
	m_resourceContainer->GetShader(L"WATER_SIMULATION_SHADER")->Render(m_deviceContext, m_Direct3D, m_resourceContainer);

	// Present the rendered scene to the screen
	m_Direct3D->EndScene();
}

void TestScene::HandleInput(Input* input, float frameTime)
{
	Camera* camera = m_resourceContainer->GetCamera(L"MAIN_CAMERA");

	// Set the frame time for calculating the updated position.
	camera->SetFrameTime(frameTime);

	// Handle the input.
	if (input->isKeyDown('W'))
	{
		// forward
		camera->MoveForward();
	}
	if (input->isKeyDown('S'))
	{
		// back
		camera->MoveBackward();
	}
	if (input->isKeyDown('A'))
	{
		// Strafe Left
		camera->StrafeLeft();
	}
	if (input->isKeyDown('D'))
	{
		// Strafe Right
		camera->StrafeRight();
	}
	if (input->isKeyDown('Q'))
	{
		// Down
		camera->MoveDownward();
	}
	if (input->isKeyDown('E'))
	{
		// Up
		camera->MoveUpward();
	}
	if (input->isKeyDown(VK_UP))
	{
		// rotate up
		camera->TurnUp();
	}
	if (input->isKeyDown(VK_DOWN))
	{
		// rotate down
		camera->TurnDown();
	}
	if (input->isKeyDown(VK_LEFT))
	{
		// rotate left
		camera->TurnLeft();
	}
	if (input->isKeyDown(VK_RIGHT))
	{
		// rotate right
		camera->TurnRight();
	}
	if (input->isKeyDown(VK_RETURN))
	{
		if (m_resourceContainer->GetFloat(L"SHADOW_SWITCH")->GetParameter(0) == 0.0f)
		{
			m_resourceContainer->GetFloat(L"SHADOW_SWITCH")->UpdateParameter(0, 1.0f);
		}
		else
		{
			m_resourceContainer->GetFloat(L"SHADOW_SWITCH")->UpdateParameter(0, 0.0f);
		}
		input->SetKeyUp(VK_RETURN);
		//CaptureGBuffer();
	}
}

void TestScene::CaptureGBuffer()
{
	ID3D11Resource* buffer;
	//m_resourceContainer->GetUAV(L"LODLookup")->GetUAV()->GetResource(&buffer);
	//SaveDDSTextureToFile(m_Direct3D->GetDeviceContext(), buffer, L"outputLOD.dds");
	m_resourceContainer->GetUAV(L"WATER_SIMULATION_01")->GetSRV()->GetResource(&buffer);
	SaveDDSTextureToFile(m_deviceContext, buffer, L"outputwater.dds");

	m_resourceContainer->GetBuffers(L"DEPTH_BUFFER")->GetShaderResourceView(0)->GetResource(&buffer);
	SaveDDSTextureToFile(m_Direct3D->GetDeviceContext(), buffer, L"outputdepth.dds");

	m_resourceContainer->GetBuffers(L"G_BUFFER")->GetShaderResourceView(0)->GetResource(&buffer);
	SaveDDSTextureToFile(m_Direct3D->GetDeviceContext(), buffer, L"outputnormal.dds");

	m_resourceContainer->GetBuffers(L"G_BUFFER")->GetShaderResourceView(1)->GetResource(&buffer);
	SaveDDSTextureToFile(m_Direct3D->GetDeviceContext(), buffer, L"outputposition.dds");

	m_resourceContainer->GetBuffers(L"LIGHT_BUFFER")->GetShaderResourceView(0)->GetResource(&buffer);
	SaveDDSTextureToFile(m_Direct3D->GetDeviceContext(), buffer, L"outputlight.dds");
	
	m_resourceContainer->GetBuffers(L"AMBIENT_OCCLUSION_TEXTURE")->GetShaderResourceView(0)->GetResource(&buffer);
	SaveDDSTextureToFile(m_Direct3D->GetDeviceContext(), buffer, L"outputao.dds");

	m_resourceContainer->GetTexture(L"RANDOM_VECTOR_MAP")->GetTexture()->GetResource(&buffer);
	SaveDDSTextureToFile(m_Direct3D->GetDeviceContext(), buffer, L"outputrandom.dds");

	m_resourceContainer->GetBuffers(L"BILATERAL_BLUR_BUFFER")->GetShaderResourceView(0)->GetResource(&buffer);
	SaveDDSTextureToFile(m_Direct3D->GetDeviceContext(), buffer, L"outputblur.dds");

	m_resourceContainer->GetLight(L"DIRECTIONAL_LIGHT")->GetShadowMap()->GetShaderResourceView(0)->GetResource(&buffer);
	SaveDDSTextureToFile(m_Direct3D->GetDeviceContext(), buffer, L"outputshadow.dds");	
	
	m_resourceContainer->GetLight(L"DIRECTIONAL_LIGHT_02")->GetShadowMap()->GetShaderResourceView(0)->GetResource(&buffer);
	SaveDDSTextureToFile(m_Direct3D->GetDeviceContext(), buffer, L"outputshadow2.dds");

	m_resourceContainer->GetBuffers(L"WATER_G_BUFFER")->GetShaderResourceView(0)->GetResource(&buffer);
	SaveDDSTextureToFile(m_Direct3D->GetDeviceContext(), buffer, L"outputwaternormal.dds");

	m_resourceContainer->GetBuffers(L"WATER_G_BUFFER")->GetShaderResourceView(1)->GetResource(&buffer);
	SaveDDSTextureToFile(m_Direct3D->GetDeviceContext(), buffer, L"outputwaterpos.dds");
}

void TestScene::InitWaterSimulation()
{
	// Set the thread counts for the water sim
	int dispatchSizeX = 16;
	int dispatchSizeZ = 16;

	int PointCountX = dispatchSizeX * 16;
	int PointCountY = dispatchSizeZ * 16;

	// Create the water mesh
	BaseMesh* mesh = 0;
	mesh = new WaterMesh(m_device, PointCountX, PointCountY);

	// Create the water object
	SceneObject* water_object = new SceneObject();
	water_object->AddMesh(m_resourceContainer->AddMesh(L"WATER_MESH", mesh));
	m_objects.push_back(water_object);

	// Create the generated water mesh
	BaseMesh* gw_mesh = new StreamOutputMesh(m_device, m_deviceContext, PointCountX * PointCountY * 6);

	// Set some initial data for the water simulation
	FluidColumnState* pData = new FluidColumnState[PointCountX*PointCountY];
	for (int j = 0; j < PointCountY; j++)
	{
		for (int i = 0; i < PointCountX; i++)
		{
			int x = i - 32;
			int y = j - 96;

			const float fFrequency = 0.1f;

			if (x*x + y*y != 0.0f)
			{
				pData[PointCountX*j + i].height = 40.0f * sinf(sqrt((float)(x*x + y*y)) * fFrequency) / (sqrt((float)(x*x + y*y)) * fFrequency);
				pData[PointCountX*j + i].flow = XMFLOAT4(0.0f, 0.0f, 0.0f, 0.0f);
			}
			else
			{
				pData[PointCountX*j + i].height = 40.0f;
				pData[PointCountX*j + i].flow = XMFLOAT4(0.0f, 0.0f, 0.0f, 0.0f);
			}
		}
	}

	D3D11_SUBRESOURCE_DATA InitialData;
	InitialData.pSysMem = pData;
	InitialData.SysMemPitch = 0;
	InitialData.SysMemSlicePitch = 0;

	// Create the two water data buffers
	UnorderedAccessView* water_sim_01 = new UnorderedAccessView();
	water_sim_01->InitStructuredBuffer(m_device, m_deviceContext, sizeof(FluidColumnState), ((dispatchSizeX * 16) * (dispatchSizeZ * 16)), &InitialData);

	UnorderedAccessView* water_sim_02 = new UnorderedAccessView();
	water_sim_02->InitStructuredBuffer(m_device, m_deviceContext, sizeof(FluidColumnState), ((dispatchSizeX * 16) * (dispatchSizeZ * 16)), NULL);

	// Create the floating point parameters required by the shader
	FloatParameter* timeFactors = new FloatParameter(4, NULL);
	float dispatchData[] =
	{
		dispatchSizeX, dispatchSizeZ, dispatchSizeX * 16, dispatchSizeZ * 16
	};
	FloatParameter* dispatchSize = new FloatParameter(4, dispatchData);

	FloatParameter* timeStep = new FloatParameter(1, NULL);
	FloatParameter* pipeArea = new FloatParameter(1, NULL);
	FloatParameter* gravitation = new FloatParameter(1, NULL);
	FloatParameter* pipeLength = new FloatParameter(1, NULL);
	FloatParameter* fluidDensity = new FloatParameter(1, NULL);
	FloatParameter* columnArea = new FloatParameter(1, NULL);
	FloatParameter* dampingFactor = new FloatParameter(1, NULL);

	// Create the water shader
	BaseShader* shader = new WaterSimulationShader(m_device, m_hwnd);
	shader->AddObject(water_object);
	shader->AddUAV(m_resourceContainer->AddUAV(L"WATER_SIMULATION_01", water_sim_01));
	shader->AddUAV(m_resourceContainer->AddUAV(L"WATER_SIMULATION_02", water_sim_02));
	shader->AddFloat(m_resourceContainer->AddFloat(L"WATER_SIM_TIME_FACTORS", timeFactors));
	shader->AddFloat(m_resourceContainer->AddFloat(L"WATER_SIM_DISPATCH_SIZE", dispatchSize));
	shader->AddFloat(m_resourceContainer->AddFloat(L"WATER_SIM_TIME_STEP", timeStep));
	shader->AddFloat(m_resourceContainer->AddFloat(L"WATER_SIM_PIPE_AREA", pipeArea));
	shader->AddFloat(m_resourceContainer->AddFloat(L"WATER_SIM_GRAVITATION", gravitation));
	shader->AddFloat(m_resourceContainer->AddFloat(L"WATER_SIM_PIPE_LENGTH", pipeLength));
	shader->AddFloat(m_resourceContainer->AddFloat(L"WATER_SIM_FLUID_DENSITY", fluidDensity));
	shader->AddFloat(m_resourceContainer->AddFloat(L"WATER_SIM_COLUMN_AREA", columnArea));
	shader->AddFloat(m_resourceContainer->AddFloat(L"WATER_SIM_DAMPING_FACTOR", dampingFactor));
	shader->AddCamera(m_resourceContainer->GetCameraIndex(L"MAIN_CAMERA"));
	m_resourceContainer->AddShader(L"WATER_SIMULATION_SHADER", shader);

	// Set the initial data for static simulation factors
	float dispatchFactors[] =
	{
		dispatchSizeX,
		dispatchSizeZ,
		dispatchSizeX * 16,
		dispatchSizeZ * 16
	};

	m_resourceContainer->GetFloat(L"WATER_SIM_DISPATCH_SIZE")->UpdateAllData(dispatchFactors);
	m_resourceContainer->GetFloat(L"WATER_SIM_PIPE_AREA")->UpdateParameter(0, 0.0002f);
	m_resourceContainer->GetFloat(L"WATER_SIM_GRAVITATION")->UpdateParameter(0, 20.0f);
	m_resourceContainer->GetFloat(L"WATER_SIM_PIPE_LENGTH")->UpdateParameter(0, 0.2f);
	m_resourceContainer->GetFloat(L"WATER_SIM_FLUID_DENSITY")->UpdateParameter(0, 1.0f);
	m_resourceContainer->GetFloat(L"WATER_SIM_COLUMN_AREA")->UpdateParameter(0, 0.05f);
	m_resourceContainer->GetFloat(L"WATER_SIM_DAMPING_FACTOR")->UpdateParameter(0, 0.9995f);

	water_object->SetScale(XMFLOAT3(32.0f / 256.0f, 32.0f / 256.0f, 32.0f / 256.0f));
	water_object->SetPosition(XMFLOAT3(0.0f, 0.1f, 0.0f));

}

void TestScene::InitTerrainGeneration()
{
	// Create the heightmap uav
	UnorderedAccessView* heightmap = new UnorderedAccessView();
	heightmap->InitRWTexture(m_device, m_deviceContext, 512, 512, NULL);

	// Create the heightmap generation shader
	BaseShader* shader = 0;
	shader = new HeightmapGenerationShader(m_device, m_hwnd);
	shader->AddUAV(m_resourceContainer->AddUAV(L"GENERATED_HEIGHTMAP", heightmap));
	m_resourceContainer->AddShader(L"HEIGHTMAP_GENERATION_SHADER", shader);

	// Run the heightmap generation shader
	shader->Render(m_deviceContext, m_Direct3D, m_resourceContainer);

	// Create the terrain mesh
	BaseMesh* mesh = 0;
	mesh = new TerrainMesh(m_device, m_deviceContext, NULL, 32, 32);

	//Create the terrain object
	SceneObject* terrain = new SceneObject();
	terrain->AddMesh(m_resourceContainer->AddMesh(L"TERRAIN_MESH", mesh));
	terrain->SetSpecularAlbedo(0.4f, 0.4f, 0.4f);
	terrain->SetSpecularPower(100.0f);
	terrain->SetScale(XMFLOAT3(32.0f, 1.0f, 32.0f));
	terrain->SetName(L"TERRAIN_GENERATION_MESH");

	m_objects.push_back(terrain);

	// Create the lod lookup uav
	UnorderedAccessView* uav = new UnorderedAccessView();
	uav->InitRWTexture(m_device, m_deviceContext, 32, 32, NULL);

	// Create the terrain shader
	shader = 0;
	shader = new TerrainShader(m_device, m_hwnd);
	shader->AddLight(m_resourceContainer->GetLightIndex(L"DIRECTIONAL_LIGHT"));
	shader->AddCamera(m_resourceContainer->GetCameraIndex(L"MAIN_CAMERA"));
	shader->AddUAV(m_resourceContainer->AddUAV(L"LODLookup", uav));
	shader->AddUAV(m_resourceContainer->GetUAVIndex(L"GENERATED_HEIGHTMAP"));
	shader->AddObject(terrain);
	shader->AddObject(GetObjectByName(L"TESSELLATED_TERRAIN"));
	m_resourceContainer->AddShader(L"TERRAIN_SHADER", shader);

	// Create the heightmap preprocess shader
	shader = 0;
	shader = new HeightmapProcessShader(m_device, m_hwnd);
	shader->AddUAV(m_resourceContainer->GetUAVIndex(L"LODLookup"));
	shader->AddUAV(m_resourceContainer->GetUAVIndex(L"GENERATED_HEIGHTMAP"));

	// Run the heightmap preprocess shader
	shader->Render(m_deviceContext, m_Direct3D, m_resourceContainer);
}

void TestScene::InitAmbientOcclusion()
{
	// Create the random vector map
	D3D11_SUBRESOURCE_DATA initData;
	initData.SysMemPitch =  256 * sizeof(XMFLOAT4);

	srand(time(NULL));

	XMFLOAT4* randomVectors = new XMFLOAT4[256 * 256];
	for (int i = 0; i < 256; ++i)
	{
		for (int j = 0; j < 256; ++j)
		{
			XMFLOAT4 random = XMFLOAT4(randf(0.0f, 1.0f), randf(0.0f, 1.0f), randf(0.0f, 1.0f), 0.0f);

			randomVectors[i * 256 + j] = random;
		}
	}

	initData.pSysMem = randomVectors;

	Texture* randomMap = new Texture(m_device, m_deviceContext, 256, 256, &initData);

	// Create the ambient occlusion texture
	DeferredBuffers* buffers = 0;
	buffers = new DeferredBuffers(m_device, m_deviceContext, m_Direct3D->GetScreenWidth() / 2, m_Direct3D->GetScreenHeight() / 2, 1, 1);

	// Create the ambient occlusion shader
	BaseShader* shader = 0;
	shader = new AmbientOcclusionShader(m_device, m_hwnd);
	shader->AddTexture(m_resourceContainer->AddTexture(L"RANDOM_VECTOR_MAP", randomMap));
	shader->AddBuffers(m_resourceContainer->GetBufferIndex(L"DEPTH_BUFFER"));
	shader->AddBuffers(m_resourceContainer->AddBuffer(L"AMBIENT_OCCLUSION_TEXTURE", buffers));
	shader->AddCamera(m_resourceContainer->GetCameraIndex(L"MAIN_CAMERA"));
	shader->AddObject(GetObjectByName(L"RENDER_WINDOW"));
	m_resourceContainer->AddShader(L"AMBIENT_OCCLUSION_SHADER", shader);

	// Create the bilateral blur buffer
	buffers = 0;
	buffers = new DeferredBuffers(m_device, m_deviceContext, m_Direct3D->GetScreenWidth() / 2, m_Direct3D->GetScreenHeight() / 2, 1, 1);

	// Create the bilateral blur shader
	shader = 0;
	shader = new BilateralBlurShader(m_device, m_hwnd);
	shader->AddBuffers(m_resourceContainer->AddBuffer(L"BILATERAL_BLUR_BUFFER", buffers));
	shader->AddBuffers(m_resourceContainer->GetBufferIndex(L"AMBIENT_OCCLUSION_TEXTURE"));
	shader->AddBuffers(m_resourceContainer->GetBufferIndex(L"DEPTH_BUFFER"));
	shader->AddObject(GetObjectByName(L"RENDER_WINDOW"));
	m_resourceContainer->AddShader(L"BILATERAL_BLUR_SHADER", shader);
}