#include "WaterSimulationShader.h"
#include "../DXFramework/ResourceContainer.h"

WaterSimulationShader::WaterSimulationShader(ID3D11Device* device, HWND hwnd) : BaseShader(device, hwnd)
{
	m_deferredShader = false;
	InitShader(L"../res/shaders/water_simulation_vs.hlsl", L"../res/shaders/water_simulation_gs.hlsl", L"../res/shaders/water_simulation_ps.hlsl", L"../res/shaders/water_simulation_cs.hlsl");
}

WaterSimulationShader::~WaterSimulationShader()
{
	if (m_dispatchParamsBuffer)
	{
		m_dispatchParamsBuffer->Release();
		m_dispatchParamsBuffer = 0;
	}

	if (m_timeParamsBuffer)
	{
		m_timeParamsBuffer->Release();
		m_timeParamsBuffer = 0;
	}

	if (m_matrixBuffer)
	{
		m_matrixBuffer->Release();
		m_matrixBuffer = 0;
	}

	if (m_simulationParamsBuffer)
	{
		m_simulationParamsBuffer->Release();
		m_simulationParamsBuffer = 0;
	}

	if (m_reflectionParamsBuffer)
	{
		m_reflectionParamsBuffer->Release();
		m_reflectionParamsBuffer = 0;
	}
}

void WaterSimulationShader::InitShader(WCHAR* vsFilename, WCHAR* psFilename)
{
}

void WaterSimulationShader::InitShader(WCHAR* vsFilename, WCHAR* gsFilename, WCHAR* psFilename, WCHAR* csFilename)
{
	// Load the shaders
	loadVertexShader(vsFilename);
	loadGeometryShader(gsFilename);
	loadPixelShader(psFilename);
	loadComputeShader(csFilename);
	loadGeometryShader(gsFilename);

	//  Load the buffers
	m_matrixBuffer = CreateConstantBuffer(m_device, sizeof(MatrixBufferType), true, true, NULL);
	m_dispatchParamsBuffer = CreateConstantBuffer(m_device, sizeof(DispatchParams), true, true, NULL);
	m_timeParamsBuffer = CreateConstantBuffer(m_device, sizeof(TimeParams), true, true, NULL);
	m_simulationParamsBuffer = CreateConstantBuffer(m_device, sizeof(SimulationParams), true, true, NULL);
	m_reflectionParamsBuffer = CreateConstantBuffer(m_device, sizeof(ReflectionParams), true, true, NULL);

	// Setup a stencil disabled depth stencil
	D3D11_DEPTH_STENCIL_DESC depthStencilDesc;

	// Set up the description of the stencil state.
	depthStencilDesc.DepthEnable = true;
	depthStencilDesc.DepthWriteMask = D3D11_DEPTH_WRITE_MASK_ALL;
	depthStencilDesc.DepthFunc = D3D11_COMPARISON_LESS;

	depthStencilDesc.StencilEnable = false;
	depthStencilDesc.StencilReadMask = 0xFF;
	depthStencilDesc.StencilWriteMask = 0xFF;

	// Stencil operations if pixel is front-facing.
	depthStencilDesc.FrontFace.StencilFailOp = D3D11_STENCIL_OP_KEEP;
	depthStencilDesc.FrontFace.StencilDepthFailOp = D3D11_STENCIL_OP_INCR;
	depthStencilDesc.FrontFace.StencilPassOp = D3D11_STENCIL_OP_REPLACE;
	depthStencilDesc.FrontFace.StencilFunc = D3D11_COMPARISON_ALWAYS;

	// Stencil operations if pixel is back-facing.
	depthStencilDesc.BackFace.StencilFailOp = D3D11_STENCIL_OP_KEEP;
	depthStencilDesc.BackFace.StencilDepthFailOp = D3D11_STENCIL_OP_DECR;
	depthStencilDesc.BackFace.StencilPassOp = D3D11_STENCIL_OP_REPLACE;
	depthStencilDesc.BackFace.StencilFunc = D3D11_COMPARISON_ALWAYS;

	// Create the depth stencil state.
	m_device->CreateDepthStencilState(&depthStencilDesc, &m_depthStencilState);

	// set up an alpha blend state
	D3D11_BLEND_DESC blendStateDescription;

	// Create an alpha enabled blend state description
	ZeroMemory(&blendStateDescription, sizeof(D3D11_BLEND_DESC));
	blendStateDescription.RenderTarget[0].BlendEnable = TRUE;
	blendStateDescription.RenderTarget[0].SrcBlend = D3D11_BLEND_SRC_ALPHA;
	blendStateDescription.RenderTarget[0].DestBlend = D3D11_BLEND_INV_SRC_ALPHA;
	blendStateDescription.RenderTarget[0].BlendOp = D3D11_BLEND_OP_ADD;
	blendStateDescription.RenderTarget[0].SrcBlendAlpha = D3D11_BLEND_ONE;
	blendStateDescription.RenderTarget[0].DestBlendAlpha = D3D11_BLEND_ZERO;
	blendStateDescription.RenderTarget[0].BlendOpAlpha = D3D11_BLEND_OP_ADD;
	blendStateDescription.RenderTarget[0].RenderTargetWriteMask = 0x0f;

	// Create the blend state using the description.
	 m_device->CreateBlendState(&blendStateDescription, &m_blendState);
}

void WaterSimulationShader::Render(ID3D11DeviceContext* deviceContext, D3D* d3d, void* resourceContainer)
{
	HRESULT result;
	D3D11_MAPPED_SUBRESOURCE mappedResource;
	MatrixBufferType* matrixBuffer;
	TimeParams* timeParamBuffer;
	DispatchParams* dispatchParamBuffer;
	SimulationParams* simulationParamBuffer;
	ReflectionParams* reflectionParamBuffer;
	XMMATRIX world, view, proj;
	int bufferNumber;

	ResourceContainer* container = reinterpret_cast<ResourceContainer*>(resourceContainer);

	// Send the time factors to the compute shader
	result = deviceContext->Map(m_timeParamsBuffer, 0, D3D11_MAP_WRITE_DISCARD, 0, &mappedResource);
	timeParamBuffer = (TimeParams*)mappedResource.pData;
	float* timeFactors = new float[4];
	container->GetFloat(m_renderFloats[0])->GetAllData(timeFactors);
	float* dispatchFactors = new float[4];
	container->GetFloat(m_renderFloats[1])->GetAllData(dispatchFactors);
	timeParamBuffer->timeFactors = XMFLOAT4(timeFactors[0], timeFactors[1], timeFactors[2], timeFactors[3]);
	timeParamBuffer->dispatchSize = XMFLOAT4(dispatchFactors[0], dispatchFactors[1], dispatchFactors[2], dispatchFactors[3]);
	deviceContext->Unmap(m_timeParamsBuffer, 0);
	bufferNumber = 0;
	deviceContext->CSSetConstantBuffers(bufferNumber, 1, &m_timeParamsBuffer);

	// Send the simulation factors to the compute shader
	result = deviceContext->Map(m_simulationParamsBuffer, 0, D3D11_MAP_WRITE_DISCARD, 0, &mappedResource);
	simulationParamBuffer = (SimulationParams*)mappedResource.pData;
	simulationParamBuffer->timeStep = container->GetFloat(m_renderFloats[2])->GetParameter(0);
	simulationParamBuffer->pipeArea = container->GetFloat(m_renderFloats[3])->GetParameter(0);
	simulationParamBuffer->gravitation = container->GetFloat(m_renderFloats[4])->GetParameter(0);
	simulationParamBuffer->pipeLength = container->GetFloat(m_renderFloats[5])->GetParameter(0);
	simulationParamBuffer->fluidDensity = container->GetFloat(m_renderFloats[6])->GetParameter(0);
	simulationParamBuffer->columnArea = container->GetFloat(m_renderFloats[7])->GetParameter(0);
	simulationParamBuffer->dampingFactor = container->GetFloat(m_renderFloats[8])->GetParameter(0);
	simulationParamBuffer->padding = 0.0f;
	deviceContext->Unmap(m_simulationParamsBuffer, 0);
	bufferNumber = 1;
	deviceContext->CSSetConstantBuffers(bufferNumber, 1, &m_simulationParamsBuffer);

	// Flip the two uavs in the compute shader
	if (m_uavSwitch)
	{
		ID3D11UnorderedAccessView* uav = container->GetUAV(m_renderUAVs[0])->GetUAV();
		UINT offset[1] = { 0 };
		deviceContext->CSSetUnorderedAccessViews(0, 1, &uav, offset);
		ID3D11ShaderResourceView* srv = container->GetUAV(m_renderUAVs[1])->GetSRV();
		deviceContext->CSSetShaderResources(0, 1, &srv);
	}
	else
	{
		ID3D11UnorderedAccessView* uav = container->GetUAV(m_renderUAVs[1])->GetUAV();
		UINT offset[1] = { 0 };
		deviceContext->CSSetUnorderedAccessViews(0, 1, &uav, offset);
		ID3D11ShaderResourceView* srv = container->GetUAV(m_renderUAVs[0])->GetSRV();
		deviceContext->CSSetShaderResources(0, 1, &srv);
	}

	deviceContext->CSSetShader(m_computeShader, NULL, 0);
	deviceContext->Dispatch(dispatchFactors[0], dispatchFactors[1], 1);

	ID3D11UnorderedAccessView* NULL_UAV = NULL;
	deviceContext->CSSetUnorderedAccessViews(0, 1, &NULL_UAV, 0);
	ID3D11ShaderResourceView* NULL_SRV = NULL;
	deviceContext->CSSetShaderResources(0, 1, &NULL_SRV);

	d3d->GetProjectionMatrix(proj);

	// Set the dispatch size parameters in the vertex shader
	result = deviceContext->Map(m_dispatchParamsBuffer, 0, D3D11_MAP_WRITE_DISCARD, 0, &mappedResource);
	dispatchParamBuffer = (DispatchParams*)mappedResource.pData;
	dispatchParamBuffer->dispatchSize = XMFLOAT4(dispatchFactors[0], dispatchFactors[1], dispatchFactors[2], dispatchFactors[3]);
	deviceContext->Unmap(m_dispatchParamsBuffer, 0);
	bufferNumber = 0;
	deviceContext->VSSetConstantBuffers(bufferNumber, 1, &m_dispatchParamsBuffer);

	// set the reflection coloring parameters in the pixel shader
	XMFLOAT3 cameraPos = container->GetCamera(m_renderCameras[0])->GetPosition();
	result = deviceContext->Map(m_reflectionParamsBuffer, 0, D3D11_MAP_WRITE_DISCARD, 0, &mappedResource);
	reflectionParamBuffer = (ReflectionParams*)mappedResource.pData;
	reflectionParamBuffer->apexColor = { 0.0f, 0.15f, 0.66f, 1.0f };
	reflectionParamBuffer->centerColor = { 0.81f, 0.38f, 0.66f, 1.0f };
	reflectionParamBuffer->cameraPos = { cameraPos.x, cameraPos.y, cameraPos.z, 1.0f };
	deviceContext->Unmap(m_reflectionParamsBuffer, 0);
	bufferNumber = 0;
	deviceContext->PSSetConstantBuffers(bufferNumber, 1, &m_reflectionParamsBuffer);

	// Set the correct access view in the vertex shader
	ID3D11ShaderResourceView* stateSrv = 0;
	if (m_uavSwitch)
		stateSrv = container->GetUAV(m_renderUAVs[0])->GetSRV();
	else
		stateSrv = container->GetUAV(m_renderUAVs[1])->GetSRV();

	deviceContext->VSSetShaderResources(0, 1, &stateSrv);

	m_uavSwitch = !m_uavSwitch;

	// Set the vertex layout
	if (m_layout)
	{
		deviceContext->IASetInputLayout(m_layout);
	}

	for (auto camera : m_renderCameras)
	{
			container->GetCamera(camera)->GetViewMatrix(view);


			world = m_renderObjects[0]->GetTransformMatrix();

			// Send the transform data to the geometry shader
			result = deviceContext->Map(m_matrixBuffer, 0, D3D11_MAP_WRITE_DISCARD, 0, &mappedResource);
			matrixBuffer = (MatrixBufferType*)mappedResource.pData;
			matrixBuffer->world = XMMatrixTranspose(world);
			matrixBuffer->view = XMMatrixTranspose(view);
			matrixBuffer->projection = XMMatrixTranspose(proj);
			deviceContext->Unmap(m_matrixBuffer, 0);
			bufferNumber = 0;
			deviceContext->GSSetConstantBuffers(bufferNumber, 1, &m_matrixBuffer);

			
			// Set the shaders
			deviceContext->VSSetShader(m_vertexShader, NULL, 0);
			deviceContext->HSSetShader(NULL, NULL, 0);
			deviceContext->DSSetShader(NULL, NULL, 0);
			deviceContext->GSSetShader(m_geometryShader, NULL, 0);
			deviceContext->PSSetShader(m_pixelShader, NULL, 0);

			for (auto mesh : m_renderObjects[0]->GetMeshes())
			{
				container->GetMesh(mesh)->SendData(deviceContext);
				m_renderType = container->GetMesh(mesh)->GetRenderType();
				m_indexCount = container->GetMesh(mesh)->GetIndexCount();

				deviceContext->OMSetBlendState(m_blendState, m_blendFactor, 0xffffffff);

				// Render the triangle.
				switch (m_renderType)
				{
				case INDEXED:
					deviceContext->DrawIndexed(m_indexCount, 0, 0);
					break;
				case AUTO:
					deviceContext->DrawAuto();
					break;
				}

				d3d->TurnOffAlphaBlending();
			}
	}
}